"""Helper script to convert colormaps from https://colorbrewer2.org into trollimage Colormap code.

The text output by this script should be copied to the ``trollimage/colormap.py``.

"""

import json
import sys
import urllib.request

JSON_URL = "https://raw.githubusercontent.com/axismaps/colorbrewer/master/export/colorbrewer.json"


def main():
    """Print python code version of trollimage Colormap objects for each colorbrewer colormap."""
    cmap_groups = _load_colormap_info_from_colorbrewer()
    _print_colormap_group(cmap_groups["seq"], "Sequential", "sequential_colormaps")
    _print_colormap_group(cmap_groups["div"], "Diverging", "diverging_colormaps")
    _print_colormap_group(cmap_groups["qual"], "Qualitative", "qualitative_colormaps",
                          normalize_values=False)


def _load_colormap_info_from_colorbrewer() -> dict[str, dict[str, list]]:
    with urllib.request.urlopen(JSON_URL) as json_file:  # nosec: B310
        colorbrewer_dict = json.load(json_file)
        cmap_groups: dict[str, dict[str, list]] = {"div": {}, "seq": {}, "qual": {}}
        for cmap_name, cmap_info in colorbrewer_dict.items():
            max_colors = max((num_colors_str for num_colors_str in cmap_info.keys() if num_colors_str != "type"),
                             key=lambda num_color_str: int(num_color_str))
            cmap_colors = cmap_info[max_colors]
            color_tuples = [rgb_color_str.replace("rgb(", "").replace(")", "").split(",")
                            for rgb_color_str in cmap_colors]
            cmap_groups[cmap_info["type"]][cmap_name.lower()] = color_tuples
    return cmap_groups


def _print_colormap_group(cmap_group: dict[str, list], human_group_name: str, group_var_name: str,
                          normalize_values: bool = True) -> None:
    print(f"# * {human_group_name} Colormaps *\n")
    for cmap_name, cmap_colors in sorted(cmap_group.items()):
        cmap_values, color_human_strings = _color_info_as_human_friendly_strings(cmap_colors, normalize_values)
        _print_single_colormap(cmap_name, cmap_values, color_human_strings)

    print(f"{group_var_name} = {{")
    for cmap_name in sorted(cmap_group.keys()):
        print(f"    \"{cmap_name}\": {cmap_name},")
    print("}\n")


def _color_info_as_human_friendly_strings(
        cmap_colors: list[tuple[int, int, int]],
        normalize_values: bool,
) -> tuple[list[str], list[tuple[str, str, str]]]:
    num_colors = len(cmap_colors)
    color_human_strings = [
        (f"{rgb_color[0]} / 255", f"{rgb_color[1]} / 255", f"{rgb_color[2]} / 255")
        for rgb_color in cmap_colors
    ]
    cmap_values = [str(color_idx) for color_idx in range(num_colors)]
    if normalize_values:
        # 0 - 1 normalized values for non-qualitative colormaps
        cmap_values = [f"{cval} / {num_colors - 1}" for cval in cmap_values]
    return cmap_values, color_human_strings


def _print_single_colormap(cmap_name: str, cmap_values: list[str], cmap_colors: list[tuple[str, str, str]]) -> None:
    cmap_pairs = [(cval, rgb_color) for cval, rgb_color in zip(cmap_values, cmap_colors)]
    print(f"{cmap_name} = Colormap(")
    for cmap_value_str, cmap_color_tuple in cmap_pairs:
        print(f"    ({cmap_value_str}, "
              f"({cmap_color_tuple[0]}, {cmap_color_tuple[1]}, {cmap_color_tuple[2]})),")
    print(")\n")


if __name__ == "__main__":
    sys.exit(main())



# cython: language_level=3, boundscheck=False, cdivision=True, wraparound=False, initializedcheck=False, nonecheck=False, cpow=True
cimport cython

from libc.math cimport cos, sin, atan2
import numpy as np
cimport numpy as np

ctypedef fused floating:
    np.float32_t
    np.float64_t


np.import_array()

# cdef extern from "numpy/npy_math.h":
#     np.float32_t NPY_NANF
#     bint npy_isnan(np.float64_t x) nogil
#     bint npy_isnan(np.float32_t x) nogil

# Function pointer type to allow for generic high-level functions
ctypedef void (*CONVERT_FUNC)(floating[:] comp1, floating[:] comp2, floating[:] comp3, floating[:, ::1] out) noexcept nogil


cdef:
    np.float32_t bintercept = 4.0 / 29  # 0.137931
    np.float32_t delta = 6.0 / 29  # 0.206896
    np.float32_t t0 = delta ** 3  # 0.008856
    np.float32_t alpha = (delta ** -2) / 3  # 7.787037
    np.float32_t third = 1.0 / 3
    np.float32_t kappa = (29.0 / 3) ** 3  # 903.3
    np.float32_t gamma = 2.2
    np.float32_t xn = 0.95047
    np.float32_t yn = 1.0
    np.float32_t zn = 1.08883
    np.float32_t denom_n = xn + (15 * yn) + (3 * zn)
    np.float32_t uprime_n = (4 * xn) / denom_n
    np.float32_t vprime_n = (9 * yn) / denom_n


    # Compile time option to use
    # sRGB companding (default, True) or simplified gamma (False)
    # sRGB companding is slightly slower but is more accurate at
    # the extreme ends of scale
    # Unit tests tuned to sRGB companding, change with caution
    bint SRGB_COMPAND = True


def rgb2lch(
        object rgba_arr,
):
    """Convert numpy RGB[A] arrays to CIE LCh_ab (Luminance, Chroma, Hue).

    See :func:`convert_colors` for more information on color spaces.

    Args:
        rgba_arr: Numpy array of RGB or RGBA colors. The array can be any
            shape as long as the channel (band) dimension is the last (-1)
            dimension. If an Alpha (A) channel is provided it is ignored.
            Values should be between 0 and 1.

    Returns: LCH_ab (l, c, h) numpy array where the last dimension represents Hue, Chroma,
        and Luminance. Hue is in radians from -pi to pi. Chroma is from 0 to
        1. Luminance is also from 0 and 1 (usually a maximum of ~0.5).

    """
    return convert_colors(rgba_arr, "rgb", "lch")


def lch2rgb(object lch_arr):
    """Convert an LCH (luminance, chroma, hue) array to RGB.

    See :func:`convert_colors` for more information on color spaces.

    Args:
        lch_arr: Numpy array of HCL values. The array can be any
            shape as long as the channel (band) dimension is the last (-1)
            dimension. Hue must be between -pi to pi. Chroma and Luminance
            should be between 0 and 1.

    Returns: RGB array where each Red, Green, and Blue channel is between 0 and 1.

    """
    return convert_colors(lch_arr, "lch", "rgb")


def convert_colors(object input_colors, str in_space, str out_space):
    """Convert from one color space to another.

    Color Spaces
    ^^^^^^^^^^^^

    * **rgb**: Red, Green, and Blue. Each channel should be in a 0 to 1
        normalized range.
    * **lch**: LCh_ab (LCH). The CIELAB Cylindrical Luminance, Chroma, and
        Hue color space. Luminance values range from about 0 to about 100.
        Chroma values range from about 0 to 120. Hue is in radians and is
        from -pi to pi.
        See the
        `wikipedia article <https://en.wikipedia.org/wiki/CIELAB_color_space#CIEHLC_cylindrical_model>`_
        for more information.
    * **lab**: CIELAB. The cartesian version of "lch". Luminance is the same
        value as LCh. The ``a*`` and ``b*`` values range from about -120 to 120.
        See the
        `wikipedia article <https://en.wikipedia.org/wiki/CIELAB_color_space>`_
        for more info.
    * **luv**: CIELUV. Luminance and a ``u*`` and ``v*`` start component. The
        luminance values range from 0 to 100. The u and v values range from
        about -200 to 200. See the
        `wikipedia article <https://en.wikipedia.org/wiki/CIELUV>`_ for more
        info.
    * **xyz**: CIE XYZ. Values range from about 0 to 1. See the
        `wikipedia article <https://en.wikipedia.org/wiki/CIE_1931_color_space>`_
        for more info.

    Args:
        input_colors: Numpy array of input colors in ``in_space`` color space.
            The array can be of any shape, but the color dimension must be the
            last dimension. Only the first three elements in the color
            dimension will be used. So if an Alpha (A) channel is provided it
            is ignored.
        in_space: String name of the color space of the input data. Can be one
            of "rgb", "lch", "lab", "luv", or "xyz".
        out_space: String name of the color space to convert to. Available
            options are the same as for ``in_space``.

    Returns:
        Numpy array with equal shape to the input, but the last dimension is
        always length 3 to match the ``out_space`` color space.

    Notes:
        This function is called by all the individual ``<space>2<space>``
        functions. This function and all color conversion functions are
        heavily based on or taken from the
        `rio-color <https://github.com/mapbox/rio-color>`_ project which is
        under an MIT license. A copy of this license is available in the
        ``trollimage`` package and root of the git repository. The majority
        of changes made to the ``rio-color`` code were to support memory views
        in a "no GIL" way and allow for 32-bit and 64-bit floating point data.

    """
    cdef object in123_arr = input_colors[..., :3]
    cdef tuple shape = in123_arr.shape
    cdef np.ndarray in123_2d = in123_arr.reshape((-1, 3))
    cdef np.ndarray out123
    if in123_arr.dtype == np.float32:
        out123 = _call_convert_func[np.float32_t](in123_2d, in_space, out_space)
    else:
        out123 = _call_convert_func[np.float64_t](in123_2d, in_space, out_space)
    return out123.reshape(shape)


cdef np.ndarray[floating, ndim=2] _call_convert_func(
        floating[:, :] in_colors, str in_space, str out_space,
):
    cdef floating[:] in1_view, in2_view, in3_view
    cdef CONVERT_FUNC conv_func = NULL
    if in_space == "rgb":
        if out_space == "lch":
            conv_func = _rgb_to_lch[floating]
        elif out_space == "lab":
            conv_func = _rgb_to_lab[floating]
        elif out_space == "luv":
            conv_func = _rgb_to_luv[floating]
        elif out_space == "xyz":
            conv_func = _rgb_to_xyz[floating]
    elif in_space == "lch":
        if out_space == "rgb":
            conv_func = _lch_to_rgb[floating]
        elif out_space == "lab":
            conv_func = _lch_to_lab[floating]
        elif out_space == "luv":
            conv_func = _lch_to_luv[floating]
        elif out_space == "xyz":
            conv_func = _lch_to_xyz[floating]
    elif in_space == "lab":
        if out_space == "rgb":
            conv_func = _lab_to_rgb[floating]
        elif out_space == "lch":
            conv_func = _lab_to_lch[floating]
        elif out_space == "luv":
            conv_func = _lab_to_luv[floating]
        elif out_space == "xyz":
            conv_func = _lab_to_xyz[floating]
    elif in_space == "luv":
        if out_space == "rgb":
            conv_func = _luv_to_rgb[floating]
        elif out_space == "lch":
            conv_func = _luv_to_lch[floating]
        elif out_space == "lab":
            conv_func = _luv_to_lab[floating]
        elif out_space == "xyz":
            conv_func = _luv_to_xyz[floating]
    elif in_space == "xyz":
        if out_space == "rgb":
            conv_func = _xyz_to_rgb[floating]
        elif out_space == "lch":
            conv_func = _xyz_to_lch[floating]
        elif out_space == "lab":
            conv_func = _xyz_to_lab[floating]
        elif out_space == "luv":
            conv_func = _xyz_to_luv[floating]

    if conv_func is NULL:
        raise ValueError("Unknown colorspace combination")

    cdef object dtype
    if floating is np.float32_t:
        dtype = np.float32
    else:
        dtype = np.float64

    in1_view = in_colors[:, 0]
    in2_view = in_colors[:, 1]
    in3_view = in_colors[:, 2]
    cdef np.ndarray[floating, ndim=2] out_colors = np.empty((in_colors.shape[0], 3), dtype=dtype)
    cdef floating[:, ::1] out_view = out_colors
    with nogil:
        conv_func(in1_view, in2_view, in3_view, out_view)
    return out_colors



cdef void _rgb_to_lab(floating[:] r_arr, floating[:] g_arr, floating[:] b_arr, floating[:, ::1] lab_arr) noexcept nogil:
    _rgb_to_xyz[floating](r_arr, g_arr, b_arr, lab_arr)
    _xyz_to_lab[floating](lab_arr[:, 0], lab_arr[:, 1], lab_arr[:, 2], lab_arr)


cdef void _rgb_to_lch(floating[:] r_arr, floating[:] g_arr, floating[:] b_arr, floating[:, ::1] lch_arr) noexcept nogil:
    _rgb_to_xyz(r_arr, g_arr, b_arr, lch_arr)
    _xyz_to_lab[floating](lch_arr[:, 0], lch_arr[:, 1], lch_arr[:, 2], lch_arr)
    _lab_to_lch[floating](lch_arr[:, 0], lch_arr[:, 1], lch_arr[:, 2], lch_arr)


cdef void _rgb_to_luv(floating[:] r_arr, floating[:] g_arr, floating[:] b_arr, floating[:, ::1] luv_arr) noexcept nogil:
    _rgb_to_xyz(r_arr, g_arr, b_arr, luv_arr)
    _xyz_to_luv[floating](luv_arr[:, 0], luv_arr[:, 1], luv_arr[:, 2], luv_arr)


cdef void _xyz_to_lch(floating[:] x_arr, floating[:] y_arr, floating[:] z_arr, floating[:, ::1] lch_arr) noexcept nogil:
    _xyz_to_lab(x_arr, y_arr, z_arr, lch_arr)
    _lab_to_lch[floating](lch_arr[:, 0], lch_arr[:, 1], lch_arr[:, 2], lch_arr)


cdef void _lab_to_rgb(floating[:] l_arr, floating[:] a_arr, floating[:] b_arr, floating[:, ::1] rgb_arr) noexcept nogil:
    _lab_to_xyz(l_arr, a_arr, b_arr, rgb_arr)
    _xyz_to_rgb[floating](rgb_arr[:, 0], rgb_arr[:, 1], rgb_arr[:, 2], rgb_arr)


cdef void _lab_to_luv(floating[:] l_arr, floating[:] a_arr, floating[:] b_arr, floating[:, ::1] luv_arr) noexcept nogil:
    _lab_to_xyz(l_arr, a_arr, b_arr, luv_arr)
    _xyz_to_luv[floating](luv_arr[:, 0], luv_arr[:, 1], luv_arr[:, 2], luv_arr)


cdef void _lch_to_xyz(floating[:] l_arr, floating[:] c_arr, floating[:] h_arr, floating[:, ::1] xyz_arr) noexcept nogil:
    _lch_to_lab(l_arr, c_arr, h_arr, xyz_arr)
    _lab_to_xyz[floating](xyz_arr[:, 0], xyz_arr[:, 1], xyz_arr[:, 2], xyz_arr)


cdef void _lch_to_rgb(floating[:] l_arr, floating[:] c_arr, floating[:] h_arr, floating[:, ::1] rgb_arr) noexcept nogil:
    _lch_to_lab(l_arr, c_arr, h_arr, rgb_arr)
    _lab_to_xyz[floating](rgb_arr[:, 0], rgb_arr[:, 1], rgb_arr[:, 2], rgb_arr)
    _xyz_to_rgb[floating](rgb_arr[:, 0], rgb_arr[:, 1], rgb_arr[:, 2], rgb_arr)


cdef void _lch_to_luv(floating[:] l_arr, floating[:] c_arr, floating[:] h_arr, floating[:, ::1] luv_arr) noexcept nogil:
    _lch_to_lab(l_arr, c_arr, h_arr, luv_arr)
    _lab_to_xyz[floating](luv_arr[:, 0], luv_arr[:, 1], luv_arr[:, 2], luv_arr)
    _xyz_to_rgb[floating](luv_arr[:, 0], luv_arr[:, 1], luv_arr[:, 2], luv_arr)


cdef void _luv_to_lab(floating[:] l_arr, floating[:] u_arr, floating[:] v_arr, floating[:, ::1] lab_arr) noexcept nogil:
    _luv_to_xyz(l_arr, u_arr, v_arr, lab_arr)
    _xyz_to_lab[floating](lab_arr[:, 0], lab_arr[:, 1], lab_arr[:, 2], lab_arr)


cdef void _luv_to_rgb(floating[:] l_arr, floating[:] u_arr, floating[:] v_arr, floating[:, ::1] rgb_arr) noexcept nogil:
    _luv_to_xyz(l_arr, u_arr, v_arr, rgb_arr)
    _xyz_to_rgb[floating](rgb_arr[:, 0], rgb_arr[:, 1], rgb_arr[:, 2], rgb_arr)


cdef void _luv_to_lch(floating[:] l_arr, floating[:] u_arr, floating[:] v_arr, floating[:, ::1] lch_arr) noexcept nogil:
    _luv_to_xyz(l_arr, u_arr, v_arr, lch_arr)
    _xyz_to_lab[floating](lch_arr[:, 0], lch_arr[:, 1], lch_arr[:, 2], lch_arr)
    _lab_to_lch[floating](lch_arr[:, 0], lch_arr[:, 1], lch_arr[:, 2], lch_arr)


# Direct colorspace conversions

cdef void _rgb_to_xyz(floating[:] red_arr, floating[:] green_arr, floating[:] blue_arr, floating[:, ::1] xyz_arr) noexcept nogil:
    cdef floating r, g, b, rl, gl, bl, x, y, z
    cdef Py_ssize_t idx

    for idx in range(red_arr.shape[0]):
        r = red_arr[idx]
        g = green_arr[idx]
        b = blue_arr[idx]

        # convert RGB to linear scale
        rl = _to_linear_rgb(r)
        gl = _to_linear_rgb(g)
        bl = _to_linear_rgb(b)

        # matrix mult for srgb->xyz,
        # includes adjustment for reference white
        x = ((rl * 0.4124564) + (gl * 0.3575761) + (bl * 0.1804375)) / xn
        y = ((rl * 0.2126729) + (gl * 0.7151522) + (bl * 0.0721750))
        z = ((rl * 0.0193339) + (gl * 0.1191920) + (bl * 0.9503041)) / zn

        xyz_arr[idx, 0] = x
        xyz_arr[idx, 1] = y
        xyz_arr[idx, 2] = z


cdef inline floating _to_linear_rgb(floating rgb_component) noexcept nogil:
    if SRGB_COMPAND:
        return _to_linear_srgb_expand(rgb_component)
    # Use "simplified sRGB"
    return rgb_component ** gamma


cdef inline floating _to_linear_srgb_expand(floating rgb_component) noexcept nogil:
    if rgb_component <= 0.04045:
        return rgb_component / 12.92
    return ((rgb_component + 0.055) / 1.055) ** 2.4


cdef void _xyz_to_lab(floating[:] x_arr, floating[:] y_arr, floating[:] z_arr, floating[:, ::1] lab) noexcept nogil:
    cdef floating x, y, z, fx, fy, fz
    cdef floating L, a, b
    cdef Py_ssize_t idx

    for idx in range(x_arr.shape[0]):
        x = x_arr[idx]
        y = y_arr[idx]
        z = z_arr[idx]

        # convert XYZ to LAB colorspace
        if x > t0:
            fx = x ** third
        else:
            fx = (alpha * x) + bintercept

        if y > t0:
            fy = y ** third
        else:
            fy = (alpha * y) + bintercept

        if z > t0:
            fz = z ** third
        else:
            fz = (alpha * z) + bintercept

        L = (116 * fy) - 16
        a = 500 * (fx - fy)
        b = 200 * (fy - fz)

        lab[idx, 0] = L
        lab[idx, 1] = a
        lab[idx, 2] = b


cdef void _lab_to_lch(floating[:] L_arr, floating[:] a_arr, floating[:] b_arr, floating[:, ::1] lch) noexcept nogil:
    cdef Py_ssize_t idx
    cdef floating c, h

    for idx in range(L_arr.shape[0]):
        lch[idx, 0] = L_arr[idx]
        # store temporary results then write output to avoid corruption
        # if the output array is the same as the input arrays
        c = ((a_arr[idx] * a_arr[idx]) + (b_arr[idx] * b_arr[idx])) ** 0.5
        h = atan2(b_arr[idx], a_arr[idx])
        lch[idx, 1] = c
        lch[idx, 2] = h


cdef void _lch_to_lab(floating[:] l_arr, floating[:] c_arr, floating[:] h_arr, floating[:, ::1] lab_arr) noexcept nogil:
    cdef floating a, b
    cdef Py_ssize_t idx

    for idx in range(l_arr.shape[0]):
        a = c_arr[idx] * cos(h_arr[idx])
        b = c_arr[idx] * sin(h_arr[idx])
        lab_arr[idx, 0] = l_arr[idx]
        lab_arr[idx, 1] = a
        lab_arr[idx, 2] = b


cdef void _lab_to_xyz(floating[:] l_arr, floating[:] a_arr, floating[:] b_arr, floating[:, ::1] xyz_arr) noexcept nogil:
    cdef floating x, y, z, L, a, b, tx, ty, tz
    cdef Py_ssize_t idx

    for idx in range(l_arr.shape[0]):
        L = l_arr[idx]
        a = a_arr[idx]
        b = b_arr[idx]

        tx = ((L + 16) / 116.0) + (a / 500.0)
        if tx > delta:
            x = tx ** 3
        else:
            x = 3 * delta * delta * (tx - bintercept)

        ty = (L + 16) / 116.0
        if ty > delta:
            y = ty ** 3
        else:
            y = 3 * delta * delta * (ty - bintercept)

        tz = ((L + 16) / 116.0) - (b / 200.0)
        if tz > delta:
            z = tz ** 3
        else:
            z = 3 * delta * delta * (tz - bintercept)

        xyz_arr[idx, 0] = x
        xyz_arr[idx, 1] = y
        xyz_arr[idx, 2] = z


cdef void _xyz_to_rgb(floating[:] x_arr, floating[:] y_arr, floating[:] z_arr, floating[:, ::1] rgb_arr) noexcept nogil:
    cdef floating rlin, glin, blin, r, g, b, x, y, z
    cdef Py_ssize_t idx

    for idx in range(x_arr.shape[0]):
        x = x_arr[idx]
        y = y_arr[idx]
        z = z_arr[idx]

        # uses reference white d65
        x = x * xn
        z = z * zn

        # XYZ to sRGB
        # expanded matrix multiplication
        rlin = (x * 3.2404542) + (y * -1.5371385) + (z * -0.4985314)
        glin = (x * -0.9692660) + (y * 1.8760108) + (z * 0.0415560)
        blin = (x * 0.0556434) + (y * -0.2040259) + (z * 1.0572252)

        r  = _to_nonlinear_rgb(rlin)
        g  = _to_nonlinear_rgb(glin)
        b  = _to_nonlinear_rgb(blin)

        # constrain to 0..1 to deal with any float drift
        if r > 1.0:
            r = 1.0
        elif r < 0.0:
            r = 0.0
        if g > 1.0:
            g = 1.0
        elif g < 0.0:
            g = 0.0
        if b > 1.0:
            b = 1.0
        elif b < 0.0:
            b = 0.0

        rgb_arr[idx, 0] = r
        rgb_arr[idx, 1] = g
        rgb_arr[idx, 2] = b


cdef inline floating _to_nonlinear_rgb(floating rgb_component) noexcept nogil:
    if SRGB_COMPAND:
        return _to_nonlinear_srgb_compand(rgb_component)
    # Use "simplified sRGB"
    return rgb_component ** (1 / gamma)


cdef inline floating _to_nonlinear_srgb_compand(floating rgb_component) noexcept nogil:
    if rgb_component <= 0.0031308:
        return 12.92 * rgb_component
    return (1.055 * (rgb_component ** (1 / 2.4))) - 0.055


cdef void _xyz_to_luv(floating[:] x_arr, floating[:] y_arr, floating[:] z_arr, floating[:, ::1] luv_arr) noexcept nogil:
    cdef floating L, u, v, uprime, vprime, denom, x, y, z
    cdef Py_ssize_t idx

    for idx in range(x_arr.shape[0]):
        x = x_arr[idx]
        y = y_arr[idx]
        z = z_arr[idx]

        denom = x + (15 * y) + (3 * z)
        uprime = (4 * x) / denom
        vprime = (9 * y) / denom

        y = y / yn

        if y <= t0:
            L = kappa * y
        else:
            L = (116 * (y ** third)) - 16

        u = 13 * L * (uprime - uprime_n)
        v = 13 * L * (vprime - vprime_n)

        luv_arr[idx, 0] = L
        luv_arr[idx, 1] = u
        luv_arr[idx, 2] = v


cdef void _luv_to_xyz(floating[:] l_arr, floating[:] u_arr, floating[:] v_arr, floating[:, ::1] xyz_arr) noexcept nogil:
    cdef floating x, y, z, uprime, vprime, L, u, v
    cdef Py_ssize_t idx

    for idx in range(l_arr.shape[0]):
        L = l_arr[idx]
        u = u_arr[idx]
        v = v_arr[idx]

        if L == 0.0:
            xyz_arr[idx, 0] = 0.0
            xyz_arr[idx, 1] = 0.0
            xyz_arr[idx, 2] = 0.0
            continue

        uprime = (u / (13 * L)) + uprime_n
        vprime = (v / (13 * L)) + vprime_n

        if L <= 8.0:
            y = L / kappa
        else:
            y = ((L + 16) / 116.0) ** 3

        x = y * ((9 * uprime) / (4 * vprime))
        z = y * ((12 - (3 * uprime) - (20 * vprime)) / (4 * vprime))

        xyz_arr[idx, 0] = x
        xyz_arr[idx, 1] = y
        xyz_arr[idx, 2] = z



"""This module defines the XRImage class.

It overlaps largely with the PIL library, but has the advantage of using
:class:`~xarray.DataArray` objects backed by :class:`dask arrays
<dask.array.Array>` as pixel arrays. This allows for invalid values to
be tracked, metadata to be assigned, and stretching to be lazy
evaluated. With the optional ``rasterio`` library installed dask array
chunks can be saved in parallel.

"""
from __future__ import annotations

import logging
import numbers
import os
import pathlib
import warnings
from collections.abc import Sequence
from typing import IO, Callable

import dask.array as da
import numpy as np
import xarray as xr
from PIL import Image as PILImage
from trollimage.image import check_image_format

logger = logging.getLogger(__name__)


def combine_scales_offsets(*args):
    """Combine ``(scale, offset)`` tuples in one, considering they are applied from left to right.

    For example, if we have our base data called ```orig_data`` and apply to it
    ``(scale_1, offset_1)``, then ``(scale_2, offset_2)`` such that::

      data_1 = orig_data * scale_1 + offset_1
      data_2 = data_1 * scale_2 + offset_2

    this function will return the tuple ``(scale, offset)`` such that::

      data_2 = orig_data * scale + offset

    given the arguments ``(scale_1, offset_1), (scale_2, offset_2)``.

    """
    cscale = 1
    coffset = 0
    for scale, offset in args:
        cscale *= scale
        coffset = coffset * scale + offset
    return cscale, coffset


def invert_scale_offset(scale, offset):
    """Invert scale and offset to allow reverse transformation.

    Ie, it will return ``rscale, roffset`` such that::

      orig_data = rscale * data + roffset

    if::

      data = scale * orig_data + offset

    """
    return 1 / scale, -offset / scale


def lazy_pil_save(
        pil_ready_arr: da.Array,
        mode: str,
        pathname_or_io: str | os.PathLike | IO[bytes],
        *args,
        **kwargs,
) -> da.Array:
    """Save PIL Image in a dask-friendly way.

    Special wrapper to handle `fill_value` try/except catch and provide a
    more useful error message.

    """
    save_res = da.map_blocks(
        _save_img_arr,
        pil_ready_arr.rechunk(pil_ready_arr.shape),
        mode,
        pathname_or_io,
        args,
        kwargs,
        chunks=((1,),),
        drop_axis=list(range(pil_ready_arr.ndim))[1:],  # drop all but the first axis
        dtype=object,
        meta=np.ndarray((), dtype=object),
    )
    return save_res


def _save_img_arr(
        pil_ready_arr: np.ndarray,
        mode: str,
        pathname_or_io: str | os.PathLike | IO[bytes],
        pil_args: tuple,
        pil_kwargs: dict,
) -> np.ndarray:
    img = PILImage.fromarray(pil_ready_arr, mode)

    try:
        img.save(pathname_or_io, *pil_args, **pil_kwargs)
    except OSError as e:
        # ex: cannot write mode LA as JPEG
        if "A as JPEG" in str(e):
            new_msg = ("Image mode not supported for this format. Specify "
                       "`fill_value=0` to set invalid values to black.")
            raise OSError(new_msg) from e
        raise

    ret_val = pil_args[0] if isinstance(pil_args[0], (str, os.PathLike)) else None
    return np.array([ret_val], dtype=object)


class XRImage:
    """Image class using an :class:`xarray.DataArray` as internal storage.

    It can be saved to a variety of image formats, but if Rasterio is
    installed, it can save to geotiff and jpeg2000 with geographical
    information.

    The enhancements functions are recording some parameters in the image's
    data attribute called `enhancement_history`.

    """

    def __init__(self, data):
        """Initialize the image with a :class:`~xarray.DataArray`."""
        data = self._correct_dims(data)

        # 'data' is an XArray, get the data from it as a dask array
        if not isinstance(data.data, da.Array):
            logger.debug("Convert image data to dask array")
            data.data = da.from_array(data.data, chunks=(data.sizes['bands'], 4096, 4096))

        self.data = data
        self.height, self.width = self.data.sizes['y'], self.data.sizes['x']
        self.palette = None

    @staticmethod
    def _correct_dims(data):
        """Standardize dimensions to bands, y, and x."""
        if not hasattr(data, 'dims'):
            raise TypeError("Data must have a 'dims' attribute.")

        if 'y' not in data.dims or 'x' not in data.dims:
            if data.ndim != 2:
                raise ValueError("Data must have a 'y' and 'x' dimension")

            # rename dimensions so we can use them
            # don't rename 'x' or 'y' if they already exist
            if 'y' not in data.dims:
                # find a dimension that isn't 'x'
                old_dim = [d for d in data.dims if d != 'x'][0]
                data = data.rename({old_dim: 'y'})
            if 'x' not in data.dims:
                # find a dimension that isn't 'y'
                old_dim = [d for d in data.dims if d != 'y'][0]
                data = data.rename({old_dim: 'x'})

        if "bands" not in data.dims:
            if data.ndim <= 2:
                data = data.expand_dims('bands')
                data['bands'] = ['L']
            else:
                raise ValueError("No 'bands' dimension provided.")
        elif "bands" not in data.coords or not np.issubdtype(data.coords["bands"].dtype, str):
            bands_by_size = {1: "L", 3: "RGB", 4: "RGBA"}
            if "mode" in data.attrs:
                mode = data.attrs["mode"]
                mode_method = "'mode' attribute"
            else:
                mode = bands_by_size[data.sizes["bands"]]
                mode_method = "dimension size"
            warnings.warn(f"Missing 'bands' coordinate. Assigning bands based on {mode_method}: {mode}",
                          stacklevel=3)
            data.coords["bands"] = list(mode)

        # doesn't actually copy the data underneath
        # we don't want our operations to change the user's data
        # we do this last in case `expand_dims` made the data read only
        data = data.copy()

        return data

    @property
    def mode(self):
        """Mode of the image."""
        return ''.join(self.data['bands'].values)

    @staticmethod
    def _gtiff_to_cog_kwargs(format_kwargs):
        """Convert GDAL Geotiff driver options to COG driver options.

        The COG driver automatically sets some format options
        but zlevel is called level and blockxsize is called blocksize.
        Convert kwargs to save() from GTiff driver to COG driver.

        """
        format_kwargs.pop('photometric', None)
        if 'zlevel' in format_kwargs:
            format_kwargs['level'] = format_kwargs.pop('zlevel')
        if 'jpeg_quality' in format_kwargs:
            format_kwargs['quality'] = format_kwargs.pop('jpeg_quality')
        format_kwargs.pop('tiled', None)
        if 'blockxsize' in format_kwargs:
            format_kwargs['blocksize'] = format_kwargs.pop('blockxsize')
        format_kwargs.pop('blockysize', None)
        return format_kwargs

    def save(self, filename, fformat=None, fill_value=None, compute=True,
             keep_palette=False, cmap=None, driver=None, **format_kwargs):
        """Save the image to the given *filename*.

        Args:
            filename (str): Output filename
            fformat (str): File format of output file (optional). Can be
                           one of many image formats supported by the
                           `rasterio` or `PIL` libraries ('jpg', 'png',
                           'tif'). By default this is determined by the
                           extension of the provided filename.
                           If the format allows, geographical information will
                           be saved to the ouput file, in the form of grid
                           mapping or ground control points.
            driver (str): can override the choice of rasterio/gdal driver
                        which is normally selected from the filename or fformat.
                        This is an implementation detail normally avoided but
                        can be necessary if you wish to distinguish between
                        GeoTIFF drivers ("GTiff" is the default, but you can
                        specify "COG" to write a Cloud-Optimized GeoTIFF).
            fill_value (float): Replace invalid data values with this value
                                and do not produce an Alpha band. Default
                                behavior is to create an alpha band.
            compute (bool): If True (default) write the data to the file
                            immediately. If False the return value is either
                            a `dask.Delayed` object or a tuple of
                            ``(source, target)`` to be passed to
                            `dask.array.store`.
            keep_palette (bool): Saves the palettized version of the image if
                                 set to True. False by default.  Warning: this
                                 does not automatically write the colormap
                                 (palette) to the file.  To write the colormap
                                 to the file, one should additionally pass the
                                 colormap with the ``cmap`` keyword argument.
            cmap (Colormap or dict): Colormap to be applied to the image when
                                     saving with rasterio, used with
                                     keep_palette=True. Should be uint8.
            format_kwargs: Additional format options to pass to `rasterio`
                           or `PIL` saving methods. Any format argument passed
                           at this stage would be superseeded by `fformat`.

        Returns:
            Either `None` or the saved filename if `compute` is True or
            ``(source, target)`` pair to be passed to `dask.array.store`.
            If compute is False the return value depends on format and how
            the image backend is used. If ``(source, target)`` is provided
            then target is an open file-like object that must be closed by
            the caller.

        """
        kwformat = format_kwargs.pop('format', None)
        fformat = fformat or kwformat or os.path.splitext(filename)[1][1:]
        if fformat in ('tif', 'tiff', 'jp2'):
            try:
                return self.rio_save(filename, fformat=fformat, driver=driver,
                                     fill_value=fill_value, compute=compute,
                                     keep_palette=keep_palette, cmap=cmap,
                                     **format_kwargs)
            except ImportError:
                logger.warning("Missing 'rasterio' dependency to save GeoTIFF "
                               "image. Will try using PIL...")
        return self.pil_save(filename, fformat, fill_value,
                             compute=compute, **format_kwargs)

    def rio_save(self, filename, fformat=None, fill_value=None,
                 dtype=np.uint8, compute=True, tags=None,
                 keep_palette=False, cmap=None, overviews=None,
                 overviews_minsize=256, overviews_resampling=None,
                 include_scale_offset_tags=False,
                 scale_offset_tags=None,
                 colormap_tag=None,
                 driver=None,
                 **format_kwargs):
        """Save the image using rasterio.

        Args:
            filename (string): The filename to save to.
            fformat (string): The format to save to. If not specified (default),
                it will be infered from the file extension.
            driver (string): The gdal driver to use. If not specified (default),
                it will be inferred from the fformat, but you can override the
                default GeoTIFF driver ("GTiff") with "COG" if you want to create
                a Cloud_Optimized GeoTIFF (and set `tiled=True,overviews=[]`).
            fill_value (number): The value to fill the missing data with.
                Default is ``None``, translating to trying to keep the data
                transparent.
            dtype (np.dtype): The type to save the data to. Defaults to
                np.uint8.
            compute (bool): Whether (default) or not to compute the lazy data.
            tags (dict): Tags to include in the file.
            keep_palette (bool): Whether or not (default) to keep the image in
                P mode.
            cmap (colormap): The colormap to use for the data.
            overviews (list): The reduction factors of the overviews to include
                in the image, eg::

                    img.rio_save('myfile.tif', overviews=[2, 4, 8, 16])

                If provided as an empty list, then levels will be
                computed as powers of two until the last level has less
                pixels than `overviews_minsize`.  If driver='COG' then use
                `overviews=[]` to get a Cloud-Optimized GeoTIFF with a correct
                set of overviews created automatically.
                Default is to not add overviews.
            overviews_minsize (int): Minimum number of pixels for the smallest
                overview size generated when `overviews` is auto-generated.
                Defaults to 256.
            overviews_resampling (str): Resampling method
                to use when generating overviews. This must be the name of an
                enum value from :class:`rasterio.enums.Resampling` and
                only takes effect if the `overviews` keyword argument is
                provided. Common values include `nearest` (default),
                `bilinear`, `average`, and many others. See the rasterio
                documentation for more information.
            scale_offset_tags (Tuple[str, str] or Dict[str, number] or None):
                If set to a ``(str, str)`` tuple, scale and offset will be
                stored in GDALMetaData tags.  Those can then be used to
                retrieve the original data values from pixel values.
                Scale and offset will be set to (NaN, NaN) for images that had
                non-linear enhancements applied (ex. gamma) as they can't be
                represented by a simple scale and offset. Scale and offset
                are also saved as (NaN, NaN) for multi-band images (ex. RGB)
                as storing multiple values in a single GDALMetaData tag is not
                currently supported.  If set to a dictionary, automatically
                determined scale/offset values are overruled by the values
                provided in the keys.
            colormap_tag (str or None):
                If set and the image was colorized or palettized, a tag will
                be added with this name with the value of a comma-separated
                version of the Colormap that was used. See
                :meth:`trollimage.colormap.Colormap.to_csv` for more
                information.

        Returns:
            The filename saved if ``compute`` is ``True``. Otherwise, a two element
            tuple to be passed to :func:`dask.array.core.store`. The first element
            represents the source or sources to save and the second element is the
            target or targets to save to. Sources and targets may be a single object
            or a list of objects. The targets may be file-like objects that must be
            closed by the user.

        """
        from ._xrimage_rasterio import RIOFile, RIODataset, split_regular_vs_lazy_tags
        fformat = fformat or os.path.splitext(filename)[1][1:]
        drivers = {'jpg': 'JPEG',
                   'png': 'PNG',
                   'tif': 'GTiff',
                   'tiff': 'GTiff',
                   'jp2': 'JP2OpenJPEG'}
        # If fformat is specified but not driver then convert it into a driver
        driver = driver or drivers.get(fformat, fformat)
        # The COG driver adds overviews so we don't need to create them ourself.
        # One thing we can't do is prevent any overviews, if we use None then
        # the COG driver will create automatically, we can't pass OVERVIEWS=NONE.
        if driver == 'COG' and overviews == []:
            overviews = None
        if include_scale_offset_tags:
            warnings.warn(
                "include_scale_offset_tags is deprecated, please use "
                "scale_offset_tags to indicate tag labels",
                DeprecationWarning, stacklevel=2)
            scale_offset_tags = scale_offset_tags or ("scale", "offset")

        if tags is None:
            tags = {}

        data, mode = self.finalize(fill_value, dtype=dtype,
                                   keep_palette=keep_palette)
        data = data.transpose('bands', 'y', 'x')

        crs = None
        gcps = None
        transform = None
        if driver in ['COG', 'GTiff', 'JP2OpenJPEG']:
            if not np.issubdtype(data.dtype, np.floating):
                format_kwargs.setdefault('compress', 'DEFLATE')
            photometric_map = {
                'RGB': 'RGB',
                'RGBA': 'RGB',
                'CMYK': 'CMYK',
                'CMYKA': 'CMYK',
                'YCBCR': 'YCBCR',
                'YCBCRA': 'YCBCR',
            }
            if mode.upper() in photometric_map:
                format_kwargs.setdefault('photometric',
                                         photometric_map[mode.upper()])

            from ._xrimage_rasterio import get_data_arr_crs_transform_gcps
            crs, transform, gcps = get_data_arr_crs_transform_gcps(data)

            stime = data.attrs.get("start_time")
            if stime:
                stime_str = stime.strftime("%Y:%m:%d %H:%M:%S")
                tags.setdefault('TIFFTAG_DATETIME', stime_str)
        if driver == 'JPEG' and 'A' in mode:
            raise ValueError('JPEG does not support alpha')

        enhancement_colormap = self._get_colormap_from_enhancement_history(data)
        if colormap_tag and enhancement_colormap is not None:
            tags[colormap_tag] = enhancement_colormap.to_csv()
        if scale_offset_tags:
            self._add_scale_offset_to_tags(scale_offset_tags, data, tags)

        # If we are changing the driver then use appropriate kwargs
        if driver == 'COG':
            format_kwargs = self._gtiff_to_cog_kwargs(format_kwargs)

        # FIXME add metadata
        r_file = RIOFile(filename, 'w', driver=driver,
                         width=data.sizes['x'], height=data.sizes['y'],
                         count=data.sizes['bands'],
                         dtype=dtype,
                         nodata=fill_value,
                         crs=crs,
                         transform=transform,
                         gcps=gcps,
                         **format_kwargs)
        r_file.open()
        if not keep_palette:
            from ._xrimage_rasterio import color_interp
            r_file.colorinterp = color_interp(data)

        if keep_palette and cmap is not None:
            if data.dtype != 'uint8':
                raise ValueError('Rasterio only supports 8-bit colormaps')
            try:
                from trollimage.colormap import Colormap
                cmap = cmap.to_rio() if isinstance(cmap, Colormap) else cmap
                r_file.rfile.write_colormap(1, cmap)
            except AttributeError:
                raise ValueError("Colormap is not formatted correctly")

        tags, da_tags = split_regular_vs_lazy_tags(tags, r_file)
        r_file.rfile.update_tags(**tags)
        r_dataset = RIODataset(r_file, overviews,
                               overviews_resampling=overviews_resampling,
                               overviews_minsize=overviews_minsize)

        to_store = ([data.data], [r_dataset])
        if da_tags:
            to_store[0].extend([tag_pair[0] for tag_pair in da_tags])
            to_store[1].extend([tag_pair[1] for tag_pair in da_tags])

        if compute:
            # write data to the file now
            da.store(*to_store)
            to_close = to_store[1]
            for item in to_close:
                item.close()
            return filename
        # provide the data object and the opened file so the caller can
        # store them when they would like. Caller is responsible for
        # closing the file
        return to_store

    @staticmethod
    def _get_colormap_from_enhancement_history(data_arr):
        for enhance_dict in reversed(data_arr.attrs.get('enhancement_history', [])):
            if "colormap" in enhance_dict:
                return enhance_dict["colormap"]
        return None

    def pil_save(
            self,
            filename: str | pathlib.Path,
            fformat: str | None = None,
            fill_value: float | int | None = None,
            compute: bool = True,
            **format_kwargs,
    ) -> da.Array | str | pathlib.Path:
        """Save the image to the given *filename* using PIL.

        For now, the compression level [0-9] is ignored, due to PIL's
        lack of support. See also :meth:`save`.

        """
        fformat = fformat or os.path.splitext(filename)[1][1:]
        fformat = check_image_format(fformat)

        if fformat == 'png':
            # Take care of GeoImage.tags (if any).
            format_kwargs['pnginfo'] = self._pngmeta()

        pil_ready_arr, mode = self.pil_array(fill_value)
        filename_arr = lazy_pil_save(pil_ready_arr, mode, filename, fformat, **format_kwargs)
        if compute:
            fn_arr = filename_arr.compute()
            return fn_arr[0]
        return filename_arr

    def _add_scale_offset_to_tags(self, scale_offset_tags, data_arr, tags):
        scale_label, offset_label = scale_offset_tags
        try:
            scale = scale_offset_tags[scale_label]
            offset = scale_offset_tags[offset_label]
        except TypeError:
            scale, offset = self.get_scaling_from_history(data_arr.attrs.get('enhancement_history', []))
        tags[scale_label], tags[offset_label] = invert_scale_offset(scale, offset)

    def get_scaling_from_history(self, history=None):
        """Merge the scales and offsets from the history.

        If ``history`` isn't provided, the history of the current image will be
        used.
        """
        if history is None:
            history = self.data.attrs.get('enhancement_history', [])
        try:
            scaling = [(item['scale'], item['offset']) for item in history]
        except KeyError as err:
            logger.debug("Can only get combine scaling from a list of linear "
                         f"scaling operations: {err}. Setting scale and offset "
                         "to (NaN, NaN).")
            return np.nan, np.nan
        scale, offset = combine_scales_offsets(*scaling)
        scale_is_not_scalar = not isinstance(scale, numbers.Number) and len(scale) != 1
        offset_is_not_scalar = not isinstance(offset, numbers.Number) and len(offset) != 1
        if scale_is_not_scalar or offset_is_not_scalar:
            logger.debug("Multi-band scale/offset tags can't be saved to "
                         "geotiff. Setting scale and offset to (NaN, NaN).")
            return np.nan, np.nan
        return scale, offset

    def apply_pil(self, fun, output_mode, pil_args=None, pil_kwargs=None, fun_args=None, fun_kwargs=None):
        """Apply a function `fun` on the pillow image corresponding to the instance of the XRImage.

        The function shall take a pil image as first argument, and is then passed fun_args and fun_kwargs.
        In addition, the current images's metadata is passed as a keyword argument called `image_metadata`.
        It is expected to return the modified pil image.
        This function returns a new XRImage instance with the modified image data.

        The pil_args and pil_kwargs are passed to the `pil_image` method of the XRImage instance.

        """
        if pil_args is None:
            pil_args = tuple()
        if pil_kwargs is None:
            pil_kwargs = dict()
        pil_ready_arr, mode = self.pil_array(*pil_args, **pil_kwargs)

        # HACK: aggdraw.Font objects cause segmentation fault in dask tokenize
        # Remove this when aggdraw is either updated to allow type(font_obj)
        # or pycoast is updated to not accept Font objects
        # See https://github.com/pytroll/pycoast/issues/43
        # The last positional argument to the _burn_overlay function in Satpy
        # is the 'overlay' dict. This could include aggdraw.Font objects so we
        # completely remove it.
        mapblocks_kwargs = {}
        if fun.__name__ == "_burn_overlay":
            from dask.base import tokenize
            from dask.utils import funcname
            func = _delayed_apply_pil
            if fun_args is None:
                fun_args = tuple()
            if fun_kwargs is None:
                fun_kwargs = dict()
            tokenize_args = (pil_ready_arr, mode, fun, fun_args[:-1], fun_kwargs,
                             self.data.attrs, self.data.dtype, output_mode)
            dask_key_name = "%s-%s" % (
                funcname(func),
                tokenize(*tokenize_args, pure=True),
            )
            mapblocks_kwargs["token"] = dask_key_name

        new_img_data = da.map_blocks(
            _delayed_apply_pil,
            pil_ready_arr.rechunk(pil_ready_arr.shape),
            mode,
            fun,
            fun_args,
            fun_kwargs,
            self.data.dtype,
            output_mode=output_mode,
            image_metadata=self.data.attrs,
            dtype=self.data.dtype,
            meta=np.ndarray((), dtype=self.data.dtype),
            chunks=((self.data.sizes["y"],), (self.data.sizes["x"],), (len(output_mode),)),
            **mapblocks_kwargs,
        )
        # new_array = self._delayed_apply_pil(fun, img_arr, fun_args, fun_kwargs,
        #                                     self.data.attrs, output_mode,
        #                                     **mapblocks_kwargs)
        new_data = xr.DataArray(new_img_data, dims=['y', 'x', 'bands'],
                                coords={'y': self.data.coords['y'],
                                        'x': self.data.coords['x'],
                                        'bands': list(output_mode)},
                                attrs=self.data.attrs)
        return XRImage(new_data)

    def _pngmeta(self):
        """Return GeoImage.tags as a PNG metadata object.

        Inspired by:
        public domain, Nick Galbreath
        http://blog.modp.com/2007/08/python-pil-and-png-metadata-take-2.html

        """
        reserved = ('interlace', 'gamma', 'dpi', 'transparency', 'aspect')

        try:
            tags = self.tags
        except AttributeError:
            tags = {}

        # Undocumented class
        from PIL import PngImagePlugin
        meta = PngImagePlugin.PngInfo()

        # Copy from tags to new dict
        for k__, v__ in tags.items():
            if k__ not in reserved:
                meta.add_text(k__, v__, 0)

        return meta

    def _create_alpha(self, data, fill_value=None):
        """Create an alpha band DataArray object.

        If `fill_value` is provided and input data is an integer type
        then it is used to determine invalid "null" pixels instead of
        xarray's `isnull` and `notnull` methods.

        The returned array is 1 where data is valid, 0 where invalid.

        """
        not_alpha = [b for b in data.coords['bands'].values if b != 'A']
        null_mask = data.sel(bands=not_alpha)
        if np.issubdtype(data.dtype, np.integer) and fill_value is not None:
            null_mask = null_mask != fill_value
        else:
            null_mask = null_mask.notnull()
        # if any of the bands are valid, we don't want transparency
        null_mask = null_mask.any(dim='bands')
        null_mask = null_mask.expand_dims('bands')
        null_mask['bands'] = ['A']
        # changes to null_mask attrs should not effect the original attrs
        # XRImage never uses them either
        null_mask.attrs = {}
        return null_mask

    def _add_alpha(self, data, alpha=None):
        """Create an alpha channel and concatenate it to the provided data.

        If ``data`` is an integer type then the alpha band will be scaled
        to use the smallest (min) value as fully transparent and the largest
        (max) value as fully opaque. If a `_FillValue` attribute is found for
        integer type data then it is used to identify null values in the data.
        Otherwise xarray's `isnull` is used.

        For float types the alpha band spans 0 to 1.

        """
        fill_value = data.attrs.get('_FillValue', None)  # integer fill value
        null_mask = alpha if alpha is not None else self._create_alpha(data, fill_value)
        # if we are using integer data, then alpha needs to be min-int to max-int
        # otherwise for floats we want 0 to 1
        if np.issubdtype(data.dtype, np.integer):
            # xarray sometimes upcasts this calculation, so cast again
            null_mask = self._scale_to_dtype(null_mask, data.dtype).astype(data.dtype)
        attrs = data.attrs.copy()
        data = xr.concat([data, null_mask], dim="bands")
        data.attrs = attrs
        return data

    def _get_dtype_scale_offset(self, dtype, fill_value):
        dinfo = np.iinfo(dtype)
        scale = dinfo.max - dinfo.min
        offset = dinfo.min
        if fill_value is not None:
            if fill_value == dinfo.min:
                # leave the lowest value for fill value only
                offset = offset + 1
                scale = scale - 1
            elif fill_value == dinfo.max:
                # leave the top value for fill value only
                scale = scale - 1
            else:
                warnings.warn(
                    "Specified fill value will overlap with valid "
                    "data. To avoid this warning specify a fill_value "
                    "that is the minimum or maximum for the data type "
                    "being saved to.", stacklevel=3)
        return scale, offset

    def _scale_to_dtype(self, data, dtype, fill_value=None):
        """Scale provided data to dtype range assuming a 0-1 range.

        Float input data is assumed to be normalized to a 0 to 1 range.
        Integer input data is not scaled, only clipped. A float output
        type is not scaled since both outputs and inputs are assumed to
        be in the 0-1 range already.

        """
        attrs = data.attrs.copy()
        if np.issubdtype(dtype, np.integer):
            if np.issubdtype(data, np.bool_):
                # convert boolean masks to floats so they can be scaled to the output integer dtype
                data = data.astype(np.float64)
            if np.issubdtype(data, np.integer):
                # preserve integer data type
                data = data.clip(np.iinfo(dtype).min, np.iinfo(dtype).max)
            else:
                # scale float data (assumed to be 0 to 1) to full integer space
                # leave room for fill value if needed
                scale, offset = self._get_dtype_scale_offset(dtype, fill_value)
                data = data.clip(0, 1) * scale + offset
                attrs.setdefault('enhancement_history', list()).append({'scale': scale, 'offset': offset})
            data = data.round()
            if fill_value is None:
                data = data.fillna(np.iinfo(dtype).min)
        data.attrs = attrs
        return data

    def _check_modes(self, modes):
        """Check that the image is in one of the given *modes*, raise an exception otherwise."""
        if not isinstance(modes, (tuple, list, set)):
            modes = [modes]
        if self.mode not in modes:
            raise ValueError("Image not in suitable mode, expected: %s, got: %s" % (modes, self.mode))

    def _from_p(self, mode):
        """Convert the image from P or PA to RGB or RGBA."""
        self._check_modes(("P", "PA"))

        if not self.palette:
            raise RuntimeError("Can't convert palettized image, missing palette.")
        pal = np.array(self.palette)
        pal = da.from_array(pal, chunks=pal.shape)

        if pal.shape[1] == 4:
            # colormap's alpha overrides data alpha
            mode = "RGBA"
            alpha = None
        elif self.mode.endswith("A"):
            # add a new/fake 'bands' dimension to the end
            alpha = self.data.sel(bands="A").data[..., None]
            mode = mode + "A" if not mode.endswith("A") else mode
        else:
            alpha = None

        flat_indexes = self.data.sel(bands='P').data.ravel().astype('int64')
        dim_sizes = ((key, val) for key, val in self.data.sizes.items() if key != 'bands')
        dims, new_shape = zip(*dim_sizes)
        dims = dims + ('bands',)
        new_shape = new_shape + (pal.shape[1],)
        new_data = pal[flat_indexes].reshape(new_shape)
        coords = dict(self.data.coords)
        coords["bands"] = list(mode)

        if alpha is not None:
            new_arr = da.concatenate((new_data, alpha), axis=-1)
            data = xr.DataArray(new_arr, coords=coords, attrs=self.data.attrs, dims=dims)
        else:
            data = xr.DataArray(new_data, coords=coords, attrs=self.data.attrs, dims=dims)

        return data

    def _l2rgb(self, mode):
        """Convert from L (black and white) to RGB."""
        self._check_modes(("L", "LA"))

        bands = ["L"] * 3
        if mode[-1] == "A":
            bands.append("A")
        data = self.data.sel(bands=bands)
        data["bands"] = list(mode)
        return data

    def convert(self, mode):
        """Convert image to *mode*."""
        if mode == self.mode:
            return self.__class__(self.data)

        if mode not in ["P", "PA", "L", "LA", "RGB", "RGBA"]:
            raise ValueError("Mode %s not recognized." % (mode))

        if mode == self.mode + "A":
            data = self._add_alpha(self.data).data
            coords = dict(self.data.coords)
            coords["bands"] = list(mode)
            data = xr.DataArray(data, coords=coords, attrs=self.data.attrs, dims=self.data.dims)
            new_img = XRImage(data)
        elif mode + "A" == self.mode:
            # Remove the alpha band from our current image
            no_alpha = self.data.sel(bands=[b for b in self.data.coords["bands"].data if b != "A"]).data
            coords = dict(self.data.coords)
            coords["bands"] = list(mode)
            data = xr.DataArray(no_alpha, coords=coords, attrs=self.data.attrs, dims=self.data.dims)
            new_img = XRImage(data)
        elif mode.endswith("A") and not self.mode.endswith("A"):
            img = self.convert(self.mode + "A")
            new_img = img.convert(mode)
        elif self.mode.endswith("A") and not mode.endswith("A"):
            img = self.convert(self.mode[:-1])
            new_img = img.convert(mode)
        else:
            cases = {
                "P": {"RGB": self._from_p},
                "PA": {"RGBA": self._from_p},
                "L": {"RGB": self._l2rgb},
                "LA": {"RGBA": self._l2rgb}
            }
            try:
                data = cases[self.mode][mode](mode)
                new_img = XRImage(data)
            except KeyError:
                raise ValueError("Conversion from %s to %s not implemented !"
                                 % (self.mode, mode))

        if self.mode.startswith('P') and new_img.mode.startswith('P'):
            # need to copy the palette
            new_img.palette = self.palette
        return new_img

    def final_mode(self, fill_value=None):
        """Get the mode of the finalized image when provided this fill_value."""
        if fill_value is None and not self.mode.endswith('A'):
            return self.mode + 'A'
        return self.mode

    def _add_alpha_and_scale(self, data, ifill, dtype):
        alpha = self._create_alpha(data, fill_value=ifill)
        data = self._scale_to_dtype(data, dtype)
        data = data.astype(dtype)
        data = self._add_alpha(data, alpha=alpha)
        return data

    def _replace_fill_value(self, data, ifill, fill_value, dtype):
        # Add fill_value after all other calculations have been done to
        # make sure it is not scaled for the data type
        if ifill is not None and fill_value is not None:
            # cast fill value to output type so we don't change data type
            fill_value = dtype(fill_value)
            # integer fields have special fill values
            data = data.where(data != ifill, dtype(fill_value))
        elif fill_value is not None:
            data = data.fillna(dtype(fill_value))

        return data

    def _get_input_fill_value(self, data):
        # if the data are integers then this fill value will be used to check for invalid values
        if np.issubdtype(data, np.integer):
            return data.attrs.get('_FillValue')
        return None

    def _scale_and_replace_fill_value(self, data, input_fill_value, fill_value,
                                      dtype, scale, replace_fill_value):
        # scale float data to the proper dtype
        # this method doesn't cast yet so that we can keep track of NULL values
        if scale:
            data = self._scale_to_dtype(data, dtype, fill_value)
        if replace_fill_value:
            data = self._replace_fill_value(data, input_fill_value, fill_value, dtype)
        return data

    def _scale_alpha_or_fill_data(self, data, fill_value, dtype, scale, alpha, replace_fill_value):
        input_fill_value = self._get_input_fill_value(data)
        needs_alpha = alpha and fill_value is None and not self.mode.endswith('A')
        if needs_alpha:
            # We don't have a fill value or an alpha, let's add an alpha
            return self._add_alpha_and_scale(data, input_fill_value, dtype)
        return self._scale_and_replace_fill_value(data, input_fill_value,
                                                  fill_value, dtype, scale,
                                                  replace_fill_value)

    def finalize(self, fill_value=None, dtype=np.uint8, keep_palette=False):
        """Finalize the image to be written to an output file.

        This adds an alpha band or fills data with a fill_value (if
        specified). It also scales float data to the output range of the
        data type (0-255 for uint8, default). For integer input data
        this method assumes the data is already scaled to the proper
        desired range. It will still fill in invalid values and add an
        alpha band if needed. Integer input data's fill value is
        determined by a special ``_FillValue`` attribute in the
        ``DataArray`` ``.attrs`` dictionary.

        Args:
            fill_value (int or float or None): Output value to use to
                represent invalid or missing pixels. By default this is
                `None` meaning an Alpha channel will be used to represent
                the invalid values; transparent for invalid, opaque
                otherwise. Some output formats do not support alpha channels
                so a ``fill_value`` must be provided. This is determined by
                the underlying library doing the writing (pillow or rasterio).
                If specified, it should be the minimum or maximum of the
                ``dtype`` (ex. 0 or 255 for uint8). Floating point image data
                is then scaled to fit the remainder of the data type space.
                Integer image data will **not** be scaled. For example, a
                ``dtype`` of ``numpy.uint8`` and a ``fill_value`` of 0 will
                result in floating-point data being scaled linearly from 1 to 255.
            dtype (numpy.dtype): Output data type to convert the current image
                data to. Default is unsigned 8-bit integer
                (:class:`numpy.uint8`).
            keep_palette (bool): Whether to convert a paletted image to RGB/A
                or not. If ``False`` (default) then ``P`` mode images will be
                converted to ``RGB`` and ``PA`` will be converted to ``RGBA``.
                If ``True``, images with mode ``P`` or ``PA`` are kept as is
                and will not be scaled in order for their index values into a
                palette to be maintained. This flag should always be ``False``
                for non-paletted images.

        """
        if keep_palette and not self.mode.startswith('P'):
            keep_palette = False

        if not keep_palette:
            finalize_kwargs = dict(
                fill_value=fill_value, dtype=dtype,
                keep_palette=keep_palette,
            )
            if self.mode == "P":
                return self.convert("RGB").finalize(**finalize_kwargs)
            if self.mode == "PA":
                return self.convert("RGBA").finalize(**finalize_kwargs)

        if np.issubdtype(dtype, np.floating) and fill_value is None:
            logger.warning("Image with floats cannot be transparent, so "
                           "setting fill_value to 0")
            fill_value = 0

        final_data = self.data.copy()
        try:
            final_data.attrs['enhancement_history'] = list(self.data.attrs['enhancement_history'])
        except KeyError:
            pass
        with xr.set_options(keep_attrs=True):
            attrs = final_data.attrs
            final_data = self._scale_alpha_or_fill_data(
                    final_data, fill_value, dtype,
                    scale=not keep_palette,
                    alpha=not keep_palette,
                    replace_fill_value=True)
            final_data = final_data.astype(dtype)
            final_data.attrs = attrs

        return final_data, ''.join(final_data['bands'].values)

    def pil_image(
            self,
            fill_value: int | float | None = None,
            compute: bool = True,
    ) -> PILImage.Image:
        """Return a PIL image from the current image.

        Args:
            fill_value (int or float): Value to use for NaN null values.
                See :meth:`~trollimage.xrimage.XRImage.finalize` for more
                info.
            compute (bool): Deprecated. The ``False`` case is no longer
                supported, use :meth:`pil_array` instead.

        Returns:
            A ``PILImage.Image`` if ``compute`` is ``True`` (default). The
            ``compute=False`` case is no longer supported.

        """
        pil_ready_arr, mode = self.pil_array(fill_value)
        if not compute:
            raise RuntimeError("Delayed PIL Image creation is no longer supported. Set 'compute=True' (default) "
                               "to get a PIL Image object back or use the 'pil_array' method to get a dask "
                               "array and image mode that are ready to be passed to 'PILImage.fromarray' "
                               "in a later dask function.")

        return PILImage.fromarray(pil_ready_arr.compute(), mode=mode)

    def pil_array(
            self,
            fill_value: int | float | None = None,
    ) -> tuple[da.Array, str]:
        """Return a PIL-ready array wrapped in a dask Array and the associated image mode.

        The underlying array is a numpy array of this image's data that has
        been finalized and squeezed to remove dimensions of size 1. The second
        return value is the expected image mode of the data as a string
        (ex. "RGB") that can be passed to ``PIL.Image.fromarray``.

        This method exists to avoid using dask Delayed functions which at the
        time of writing result in duplicate computation of Array tasks that
        are used as input to the Delayed function. This method should be used
        when wanting to perform a lazy evaluated dask operation on a PIL Image
        object. The lazy operation (ex. a function called from dask's
        ``map_blocks``) is expected to do
        ``PIL.Image.fromarray(pil_array, mode=mode)`` with the input
        ``pil_array``.

        The dask Array will have whatever chunking this `XRImage`'s data has
        and is not guaranteed to be one large chunk. One chunk may be needed
        to process the data as one single PIL Image. To rechunk the data
        for a call to dask's ``map_blocks`` pass the rechunked version of the
        data with ``pil_ready_arr.rechunk(pil_ready_arr.shape)``.

        Args:
            fill_value (int or float): Value to use for NaN null values.
                See :meth:`~trollimage.xrimage.XRImage.finalize` for more
                info.

        Returns:
            Dask array wrapping a finalized image numpy array and the image
            mode as a string (ex. "RGB").

        """
        channels, mode = self.finalize(fill_value)
        pil_ready_arr = np.squeeze(channels.transpose('y', 'x', 'bands').data)
        return pil_ready_arr, mode

    def xrify_tuples(self, tup):
        """Make xarray.DataArray from tuple."""
        return xr.DataArray(tup,
                            dims=['bands'],
                            coords={'bands': self.data['bands']})

    def gamma(self, gamma=None):
        """Apply gamma correction to the channels of the image.

        If *gamma* is a tuple, then it should have as many elements as
        the channels of the image, and the gamma correction is applied
        elementwise. If *gamma* is a number, the same gamma correction
        is applied on every channel, if there are several channels in
        the image. The behaviour of :func:`gamma` is undefined outside
        the normal [0,1] range of the channels.

        """
        if _is_unity_or_none(gamma):
            return

        inverse_gamma = self._get_inverse_gamma(gamma)
        logger.debug("Applying gamma %s", str(gamma))
        attrs = self.data.attrs
        self.data = self.data.clip(min=0)
        self.data **= inverse_gamma
        self.data.attrs = attrs
        self.data.attrs.setdefault('enhancement_history', []).append({'gamma': gamma})

    def _get_inverse_gamma(self, gamma):
        if np.issubdtype(self.data.dtype, np.floating):
            dtype = self.data.dtype
        else:
            dtype = np.float32
        if isinstance(gamma, (list, tuple)):
            gamma = self.xrify_tuples(gamma).astype(dtype)
        else:
            gamma = np.array(gamma, dtype=dtype)
        return 1.0 / gamma

    def stretch(self, stretch="crude", **kwargs):
        """Apply stretching to the current image.

        The value of *stretch* sets the type of stretching applied. The
        values "histogram", "linear", "crude" (or "crude-stretch")
        perform respectively histogram equalization, contrast stretching
        (with 5% cutoff on both sides), and contrast stretching without
        cutoff. The value "logarithmic" or "log" will do a logarithmic
        enhancement towards white. If a tuple or a list of two values is
        given as input, then a contrast stretching is performed with the
        values as cutoff. These values should be normalized in the range
        [0.0,1.0].

        """
        logger.debug("Applying stretch %s with parameters %s",
                     stretch, str(kwargs))

        # FIXME: do not apply stretch to alpha channel

        if isinstance(stretch, (tuple, list)):
            if len(stretch) == 2:
                self.stretch_linear(cutoffs=stretch)
            else:
                raise ValueError(
                    "Stretch tuple must have exactly two elements")
        elif stretch == "linear":
            self.stretch_linear(**kwargs)
        elif stretch == "histogram":
            self.stretch_hist_equalize(**kwargs)
        elif stretch in ["crude", "crude-stretch"]:
            self.crude_stretch(**kwargs)
        elif stretch in ["log", "logarithmic"]:
            self.stretch_logarithmic(**kwargs)
        elif stretch == "no":
            return
        elif isinstance(stretch, str):
            raise ValueError("Stretching method %s not recognized." % stretch)
        else:
            raise TypeError("Stretch parameter must be a string or a tuple.")

    def stretch_linear(self, cutoffs=(0.005, 0.005)):
        """Stretch linearly the contrast of the current image.

        Use *cutoffs* for left and right trimming.

        If the cutoffs are just a tuple or list of two scalars, all the
        channels except the alpha channel will be stretched with the cutoffs.
        If the cutoffs are a sequence of tuples/lists of two scalars then:

        - if there are the same number of tuples/lists as channels, each channel will be stretched with the respective
          cutoff.
        - if there is one less tuple/list as channels, the same applies, except for the alpha channel which will
          not be stretched.

        """
        logger.debug("Perform a linear contrast stretch.")

        left, right = self._get_left_and_right_quantiles_for_linear_stretch(cutoffs)

        self.crude_stretch(left, right)

    def _get_left_and_right_quantiles_for_linear_stretch(self, cutoffs):
        logger.debug("Calculate the histogram quantiles: ")
        logger.debug("Left and right quantiles: " +
                     str(cutoffs[0]) + " " + str(cutoffs[1]))
        data = self.data
        nb_bands = len(data.coords["bands"])

        dont_stretch_alpha = ('A' in self.data.coords['bands'].values and
                              (np.isscalar(cutoffs[0]) or len(cutoffs) == nb_bands - 1))

        if np.isscalar(cutoffs[0]):
            cutoffs = [cutoffs] * nb_bands

        if dont_stretch_alpha:
            data = self.data.sel(bands=self.data.coords['bands'].values[:-1])

        left_data, right_data = self._get_left_and_right_quantiles_without_alpha(data, cutoffs)

        if dont_stretch_alpha:
            left_data = left_data + [da.zeros_like(left_data[0])]
            right_data = right_data + [da.ones_like(right_data[0])]
        left = xr.DataArray(da.stack(left_data), dims=('bands',),
                            coords={'bands': self.data['bands']})
        right = xr.DataArray(da.stack(right_data), dims=('bands',),
                             coords={'bands': self.data['bands']})
        return left, right

    @staticmethod
    def _get_left_and_right_quantiles_without_alpha(
            data_arr: xr.DataArray,
            cutoffs: Sequence[Sequence[float]],  # two-element sequences [left_cutoff, right_cutoff]
    ) -> tuple[list[da.Array], list[da.Array]]:
        left = []
        right = []
        for i in range(data_arr.sizes["bands"]):
            left_i, right_i = data_arr.isel(bands=i).quantile([cutoffs[i][0], 1-cutoffs[i][1]], dim=("y", "x"))
            left.append(left_i.data)
            right.append(right_i.data)
        return left, right

    def crude_stretch(self, min_stretch=None, max_stretch=None):
        """Perform simple linear stretching.

        This is done without any cutoff on the current image and
        normalizes to the [0,1] range.

        """
        min_stretch = self._check_stretch_value(min_stretch, kind='min')
        max_stretch = self._check_stretch_value(max_stretch, kind='max')
        scale_factor = self._get_scale_factor(min_stretch, max_stretch)

        attrs = self.data.attrs
        offset = -min_stretch * scale_factor
        try:
            offset = offset.astype(scale_factor.dtype)
        except AttributeError:
            offset = scale_factor.dtype.type(offset)

        self.data = np.multiply(self.data, scale_factor, dtype=scale_factor.dtype) + offset
        self.data.attrs = attrs
        self.data.attrs.setdefault('enhancement_history', []).append({'scale': scale_factor,
                                                                      'offset': offset})

    def _check_stretch_value(self, val, kind='min'):
        if val is None:
            non_band_dims = tuple(x for x in self.data.dims if x != 'bands')
            val = getattr(self.data, kind)(dim=non_band_dims)
        if isinstance(val, (list, tuple)):
            val = self.xrify_tuples(val)

        dtype = self.data.dtype
        if dtype in (np.uint8, np.int8, np.uint16, np.int16):
            dtype = np.dtype(np.float32)
        elif np.issubdtype(dtype, np.integer) or isinstance(dtype, int):
            dtype = np.dtype(np.float64)
        try:
            val = val.astype(dtype)
        except AttributeError:
            val = dtype.type(val)

        return val

    def _get_scale_factor(self, min_stretch, max_stretch):
        delta = (max_stretch - min_stretch)
        dtype = self._infer_scale_factor_dtype()
        if isinstance(delta, xr.DataArray):
            # fillna if delta is NaN
            scale_factor = (1.0 / delta).fillna(0).astype(dtype)
        else:
            scale_factor = np.array(1.0 / delta, dtype=dtype)

        return scale_factor

    def _infer_scale_factor_dtype(self):
        if np.issubdtype(self.data.dtype, np.integer):
            return np.float32
        return self.data.dtype

    def stretch_hist_equalize(self, approximate=False):
        """Stretch the current image's colors through histogram equalization.

        Args:
            approximate (bool): Deprecated.
        """
        logger.info("Perform a histogram equalized contrast stretch.")

        if approximate:
            warnings.warn("The 'approximate' keyword is deprecated.", stacklevel=2)

        nwidth = 2048.
        cdf = np.arange(0., 1., 1. / nwidth)

        def _band_hist(band_data: da.Array) -> da.Array:
            one_chunk_data = band_data.rechunk((-1,) * band_data.ndim)
            # FUTURE: 2025-06-24 - The next release of dask should fix the need for the `axis` keyword argument here
            # See https://github.com/dask/dask/pull/11988
            # Without the single chunk of data above, `axis` is needed, but we use the single chunk
            # during the call to `interp` so we do it here instead of letting dask do it for us
            bins = da.nanpercentile(one_chunk_data, cdf * 100.0, axis=tuple(range(band_data.ndim)))
            res = da.map_blocks(
                np.interp,
                one_chunk_data,
                bins,
                cdf,
                chunks=one_chunk_data.chunks,
                meta=np.array((), dtype=one_chunk_data.dtype),
                dtype=one_chunk_data.dtype,
            )
            return res.rechunk(band_data.chunks)

        band_results = []
        for band in self.data['bands'].values:
            if band == 'A':
                continue
            band_data = self.data.sel(bands=band)
            res = _band_hist(band_data.data)
            band_results.append(res)

        if 'A' in self.data.coords['bands'].values:
            band_results.append(self.data.sel(bands='A'))
        self.data.data = da.stack(band_results,
                                  axis=self.data.dims.index('bands'))
        self.data.attrs.setdefault('enhancement_history', []).append({'hist_equalize': True})

    def stretch_logarithmic(self, factor=100., base="e", min_stretch=None, max_stretch=None):
        """Move data into range [1:factor] through normalized logarithm.

        Args:
            factor (float): Maximum of the range data will be scaled to
                before applying the log function. Image data will be scaled
                to a 1 to ``factor`` range.
            base (str): Type of log to use. Defaults to natural log ("e"),
                but can also be "10" for base 10 log or "2" for base 2 log.
            min_stretch (float or list): Minimum input value to scale from.
                Data will be clipped to this value before being scaled to
                the 1:factor range. By default (None), the limits are computed
                on the fly but with a performance penalty. May also be a list
                for multi-band images.
            max_stretch (float or list): Maximum input value to scale from.
                Data will be clipped to this value before being scaled to
                the 1:factor range. By default (None), the limits are computed
                on the fly but with a performance penalty. May also be a list
                for multi-band images.

        """
        logger.debug("Perform a logarithmic contrast stretch.")
        crange = (0., 1.0)
        log_func = np.log if base == "e" else getattr(np, "log" + base)
        min_stretch, max_stretch = self._convert_log_minmax_stretch(min_stretch, max_stretch)

        b__ = float(crange[1] - crange[0]) / self.data.dtype.type(log_func(factor))
        c__ = float(crange[0])

        def _band_log(arr, min_input, max_input):
            slope = (factor - 1.) / (max_input - min_input)
            arr = np.clip(arr, min_input, max_input)
            arr = 1. + (arr - min_input) * slope
            arr = c__ + b__ * log_func(arr)
            return arr

        band_results = []
        for band_idx, band in enumerate(self.data['bands'].values):
            if band == 'A':
                continue
            band_data = self.data.sel(bands=band)
            res = _band_log(band_data.data,
                            min_stretch[band_idx],
                            max_stretch[band_idx])
            band_results.append(res)

        if 'A' in self.data.coords['bands'].values:
            band_results.append(self.data.sel(bands='A'))
        self.data.data = da.stack(band_results, axis=self.data.dims.index('bands'))
        self.data.attrs.setdefault('enhancement_history', []).append({'log_factor': factor})

    def _convert_log_minmax_stretch(self, min_stretch, max_stretch):
        non_band_dims = tuple(x for x in self.data.dims if x != 'bands')
        if min_stretch is None:
            min_stretch = [m.data for m in self.data.min(dim=non_band_dims)]
        if max_stretch is None:
            max_stretch = [m.data for m in self.data.max(dim=non_band_dims)]
        if not isinstance(min_stretch, (list, tuple)):
            min_stretch = [min_stretch] * self.data.sizes.get("bands", 1)
        if not isinstance(max_stretch, (list, tuple)):
            max_stretch = [max_stretch] * self.data.sizes.get("bands", 1)
        return min_stretch, max_stretch

    def stretch_weber_fechner(self, k, s0):
        """Stretch according to the Weber-Fechner law.

        p = k.ln(S/S0)
        p is perception, S is the stimulus, S0 is the stimulus threshold (the
        highest unperceived stimulus), and k is the factor.

        """
        attrs = self.data.attrs
        clipped_stimuli = np.clip(self.data, s0, None)
        self.data = k * np.log(clipped_stimuli / s0)
        self.data.attrs = attrs
        self.data.attrs.setdefault('enhancement_history', []).append({'weber_fechner': (k, s0)})

    def invert(self, invert=True):
        """Inverts all the channels of a image according to *invert*.

        If invert is a tuple or a list, elementwise invertion is performed,
        otherwise all channels are inverted if *invert* is true (default).

        Note: 'Inverting' means that black becomes white, and vice-versa, not
        that the values are negated !

        """
        logger.debug("Applying invert with parameters %s", str(invert))
        if isinstance(invert, (tuple, list)):
            invert = self.xrify_tuples(invert)
            offset = invert.astype(self.data.dtype)
            scale = (-1) ** offset
        elif invert:
            offset = 1
            scale = -1
        attrs = self.data.attrs
        self.data = self.data * scale + offset
        self.data.attrs = attrs
        self.data.attrs.setdefault('enhancement_history', []).append({'scale': scale,
                                                                      'offset': offset})

    def stack(self, img):
        """Stack the provided image on top of the current image."""
        # TODO: Conversions between different modes with notification
        # to the user, i.e. proper logging
        if self.mode != img.mode:
            raise NotImplementedError("Cannot stack images of different modes.")

        self.data = self.data.where(img.data.isnull(), img.data)

    def merge(self, img):
        """Use the provided image as background for the current *img* image.

        That is if the current image has missing data.

        """
        raise NotImplementedError("This method has not be implemented for "
                                  "xarray support.")
        if self.is_empty():
            raise ValueError("Cannot merge an empty image.")

        if self.mode != img.mode:
            raise ValueError("Cannot merge image of different modes.")

        selfmask = self.channels[0].mask
        for chn in self.channels[1:]:
            selfmask = np.ma.mask_or(selfmask, chn.mask)

        for i in range(len(self.channels)):
            self.channels[i] = np.ma.where(selfmask,
                                           img.channels[i],
                                           self.channels[i])
            self.channels[i].mask = np.logical_and(selfmask,
                                                   img.channels[i].mask)

    def colorize(self, colormap):
        """Colorize the current image using ``colormap``.

        Convert a greyscale image (mode "L" or "LA") to a color image (mode
        "RGB" or "RGBA") by applying a colormap. If floating point data being
        colorized contains NaNs then the result will also contain NaNs instead
        of a color from the colormap. Integer data that includes
        a ``.attrs['_FillValue']`` will be converted to a floating point array
        and values equal to ``_FillValue`` replaced with NaN before being
        colorized.

        To create a color image in mode "P" or "PA", use
        :meth:`~XRImage.palettize`.

        Args:
            colormap (:class:`~trollimage.colormap.Colormap`):
                Colormap to be applied to the image.

        .. note::

            Works only on "L" or "LA" images.

        """
        if self.mode not in ("L", "LA"):
            raise ValueError("Image should be grayscale to colorize")

        colormap = self._adjust_colormap_dtype(colormap)
        l_data = self._get_masked_floating_luminance_data()
        alpha = self.data.sel(bands=['A']) if self.mode == "LA" else None
        new_data = colormap.colorize(l_data.data)

        if colormap.colors.shape[1] == 4:
            mode = "RGBA"
        elif alpha is not None:
            new_data = da.concatenate([new_data, alpha.data], axis=0)
            mode = "RGBA"
        else:
            mode = "RGB"

        # copy the coordinates so we don't affect the original
        coords = dict(self.data.coords)
        coords['bands'] = list(mode)
        attrs = self.data.attrs
        dims = self.data.dims
        self.data = xr.DataArray(new_data, coords=coords, attrs=attrs, dims=dims)
        scale_factor, offset = self._get_colormap_scale_offset(colormap)
        self.data.attrs.setdefault('enhancement_history', []).append({
            'scale': scale_factor,
            'offset': offset,
            'colormap': colormap,
        })

    def _adjust_colormap_dtype(self, colormap):
        if np.issubdtype(self.data.dtype, np.floating) and colormap.colors.dtype != self.data.dtype:
            colormap.colors = colormap.colors.astype(self.data.dtype)
            colormap.values = colormap.values.astype(self.data.dtype)
        return colormap

    def _get_masked_floating_luminance_data(self):
        l_data = self.data.sel(bands='L')
        # mask any integer fields with _FillValue
        # assume NaN is used otherwise
        if self.mode == "L" and np.issubdtype(self.data.dtype, np.integer):
            fill_value = self._get_input_fill_value(self.data)
            if fill_value is not None:
                l_data = l_data.where(l_data != fill_value)
        return l_data

    def palettize(self, colormap):
        """Palettize the current image using ``colormap``.

        Convert a mode "L" (or "LA") grayscale image to a mode "P" (or "PA")
        palette image and store the palette in the ``palette`` attribute.
        To store this image in mode "P", call :meth:`~XRImage.save` with
        ``keep_palette=True``.  To include color information in the output
        format (if supported), call :meth:`~XRImage.save` with
        ``keep_palette=True`` *and* ``cmap=colormap``.

        To (directly) get an image in mode "RGB" or "RGBA", use
        :meth:`~XRImage.colorize`.

        Invalid data (NaN) are sorted into the final bin and the ``_FillValue``
        attributed for the resulting image is set to the corresponding value.

        Args:
            colormap (:class:`~trollimage.colormap.Colormap`):
                Colormap to be applied to the image.

        Notes:
            Works only on "L" or "LA" images.

            Similar to other enhancement methods (colorize, stretch, etc)
            this method adds an ``enhancement_history`` list to the metadata
            stored in the image ``DataArray``'s metadata (``.attrs``).
            In other methods, however, the metadata directly translates to
            the linear operations performed in that enhancement. The palettize
            operation converts data values to indices into a colormap.
            This result is based on the range of values defined in the Colormap
            (``cmap.values``). To be most useful, the enhancement history scale
            and offset values represent the range of the colormap as if scaling
            the data to a 0-1 range. This means that once the data is saved to
            a format as an RGB (the palette colors are applied) the scale and
            offset can be used to determine the original range of the data
            based on the min/max of the data type of the format (ex. uint8).
            For example:

            .. code-block:: python

                dtype_min = 0
                dtype_max = 255
                scale = ...  # scale from geotiff
                offset = ...  # offset from geotiff
                data_min = offset
                data_max = (dtype_max - dtype_min) * scale + offset

            If a geotiff is saved with ``keep_palette=True`` then the data
            saved to the geotiff are the palette indices and will not be
            scaled to the data type of the format. There will also be a
            standard geotiff color table in the geotiff to identify that
            these are indices rather than some other type of image data. This
            means in this case the scale and offset can be used to determine
            the original range of the data starting from a 0-1 range
            (``dtype_min`` is 0 and ``dtype_max`` is 1 in the code above).

        """
        if self.mode not in ("L", "LA"):
            raise ValueError("Image should be grayscale to colorize")

        l_data = self.data.sel(bands=['L'])
        new_data, self.palette = colormap.palettize(l_data.data)

        if self.mode == "L":
            mode = "P"
        else:
            mode = "PA"
            new_data = da.concatenate([new_data, self.data.sel(bands=['A'])], axis=0)

        old_dtype = self.data.dtype
        self.data.data = new_data
        self.data.coords['bands'] = list(mode)
        self._set_new_fill_value_after_palettize(colormap, old_dtype)
        # See docstring notes above for how scale/offset should be used
        scale_factor, offset = self._get_colormap_scale_offset(colormap)
        self.data.attrs.setdefault('enhancement_history', []).append({
            'scale': scale_factor,
            'offset': offset,
            'colormap': colormap,
        })

    @staticmethod
    def _get_colormap_scale_offset(colormap):
        cmap_min = colormap.values[0]
        cmap_max = colormap.values[-1]
        scale_factor = 1.0 / (cmap_max - cmap_min)
        offset = -cmap_min * scale_factor
        return scale_factor, offset

    def _set_new_fill_value_after_palettize(self, colormap, old_dtype):
        """Set new fill value after palettizing."""
        # OK: float without fill value or fill value nan
        # OK: int without fill value or fill value to max value
        if ((np.issubdtype(old_dtype, np.inexact) and
             np.isnan(self.data.attrs.get("_FillValue", np.nan))) or
            (np.issubdtype(old_dtype, np.integer) and
             self.data.attrs.get("_FillValue", np.iinfo(old_dtype).max) == np.iinfo(old_dtype).max)):
            self.data.attrs["_FillValue"] = colormap.values.shape[0]-1
        # not OK: float or int with different fill value
        elif "_FillValue" in self.data.attrs:
            warnings.warn(
                f"Palettizing {old_dtype.name:s} data with the _FillValue attribute set to "
                f"{self.data.attrs['_FillValue']!s}, "
                "but palettize is not generally fill value aware (masked data "
                "will be correctly palettized only for float with NaN or for "
                "ints with fill value set to dtype max.",
                UserWarning,
                stacklevel=3)
        # else: non-numeric data, probably doesn't work at all and will fail
        # elsewhere anyway

    def blend(self, src):
        r"""Alpha blend *src* on top of the current image.

        Perform `alpha blending`_ of *src* on top of the current image.
        Alpha blending is defined as:

        .. math::

           \begin{cases}
            \mathrm{out}_A =
             \mathrm{src}_A + \mathrm{dst}_A (1 - \mathrm{src}_A) \\
            \mathrm{out}_{RGB} =
             \bigl(\mathrm{src}_{RGB}\mathrm{src}_A
                 + \mathrm{dst}_{RGB} \mathrm{dst}_A
                   \left(1 - \mathrm{src}_A \right) \bigr)
             \div \mathrm{out}_A \\
            \mathrm{out}_A = 0 \Rightarrow \mathrm{out}_{RGB} = 0
           \end{cases}

        Both images must have mode ``"RGBA"``.

        Args:
            src (:class:`XRImage` with mode ``"RGBA"``)
                Image to be blended on top of current image.

        .. _alpha blending: https://en.wikipedia.org/w/index.php?title=Alpha_compositing&oldid=891033105#Alpha_blending

        Returns
            XRImage with mode "RGBA", blended as described above

        """
        # NB: docstring maths copy-pasta from enwiki

        if self.mode != "RGBA":
            raise ValueError(
                    "Expected self.mode='RGBA', got {md!s}".format(
                        md=self.mode))
        if not isinstance(src, XRImage):
            raise TypeError("Expected XRImage, got {tp!s}".format(
                tp=type(src)))
        if src.mode != "RGBA":
            raise ValueError("Expected src.mode='RGBA', got {sm!s}".format(
                sm=src.mode))

        srca = src.data.sel(bands="A")
        dsta = self.data.sel(bands="A")
        outa = srca + dsta * (1-srca)
        bi = {"bands": ["R", "G", "B"]}
        rgb = ((src.data.loc[bi] * srca
               + self.data.loc[bi] * dsta * (1-srca))
               / outa).where(outa != 0, 0)
        return self.__class__(
                xr.concat(
                    [rgb, outa.expand_dims("bands")],
                    dim="bands"))

    def show(self):
        """Display the image on screen."""
        self.pil_image().show()

    def _repr_png_(self):
        import io
        b = io.BytesIO()
        self.pil_image().save(b, format='png')
        return b.getvalue()


def _delayed_apply_pil(
        pil_ready_array: np.ndarray,
        mode: str,
        fun: Callable,
        fun_args: tuple | None,
        fun_kwargs: dict | None,
        dtype: np.dtype,
        image_metadata: dict | None = None,
        output_mode: str | None = None,
) -> np.ndarray:
    if fun_args is None:
        fun_args = tuple()
    if fun_kwargs is None:
        fun_kwargs = dict()
    if image_metadata is None:
        image_metadata = dict()
    pil_image = PILImage.fromarray(pil_ready_array, mode=mode)
    new_img = fun(pil_image, image_metadata, *fun_args, **fun_kwargs)
    if output_mode is not None:
        new_img = new_img.convert(output_mode)
    return np.array(new_img) / dtype.type(255.0)


def _is_unity_or_none(gamma):
    if gamma is None or gamma == 1.0:
        return True
    if not hasattr(gamma, "__iter__"):
        return False
    return all(g == 1.0 for g in gamma) or all(g is None for g in gamma)



"""A simple colormap module."""

import contextlib
import copy
import os
from io import StringIO
from typing import Optional
import warnings
from numbers import Number
import pathlib
import sys

import numpy as np
from trollimage.colorspaces import rgb2lch, lch2rgb


@contextlib.contextmanager
def _file_or_stringio(filename_or_none):
    if filename_or_none is None:
        yield StringIO()
    else:
        with open(filename_or_none, "w") as file_obj:
            yield file_obj


def colorize(arr, colors, values):
    """Colorize a monochromatic array *arr*, based *colors* given for *values*.

    Args:
        arr (numpy array, numpy masked array, dask array):
            Data to be mapped to the colors in the colors array using values
            as control points. Data can be any shape, but must represent a
            single (luminance) band of data (not RGB or any other colorspace).
        colors (numpy array):
            Colors to map the data to. Colors can be RGB or RGBA in the
            0 to 1 range. The array should be in the shape (N, 3 or 4) where
            N is the size of the colormap (the number of colors) and the last
            dimension is the band dimension where each element represents
            Red (R), Green (G), Blue (B), and optionally Alpha (A).
        values (numpy array):
            Control points mapping input data values to the colors to be
            mapped to. Should be one dimension and the same number of elements
            as the ``colors`` array has colors (N).

    Returns: Resulting RGB/A array with the shape (3 or 4, ...) where the
        first dimension is 3 if the colors array was RGB and 4 if the colors
        array was RGBA. The remaining shape of the result matches the provided
        ``data`` input shape. Like ``colors`` the color values will be between
        0 and 1.

    """
    if can_be_block_mapped(arr):
        return _colorize_dask(arr, colors, values)
    else:
        return _colorize(arr, colors, values)


def can_be_block_mapped(data):
    """Check if the array can be processed in chunks."""
    return hasattr(data, 'map_blocks')


def _colorize_dask(dask_array, colors, values):
    """Colorize a dask array.

    The channels are stacked on the first dimension.
    """
    return dask_array.map_blocks(_colorize, colors, values, dtype=colors.dtype, new_axis=0,
                                 chunks=[colors.shape[1]] + list(dask_array.chunks))


def _colorize(arr, colors, values):
    """Colorize the array."""
    channels = _interpolate_rgb_colors(arr, colors, values)
    alpha = _interpolate_alpha(arr, colors, values)
    channels.extend(alpha)
    channels = _mask_channels(channels, arr)
    return np.stack(channels, axis=0)


def _interpolate_rgb_colors(arr, colors, values):
    interp_xp_coords = np.array(values)
    interp_y_coords = rgb2lch(colors)
    if values[0] > values[-1]:
        # monotonically decreasing
        interp_xp_coords = interp_xp_coords[::-1]
        interp_y_coords = interp_y_coords[::-1]
    # Make sure hue (radians) are consistently increasing or decreasing
    interp_lch = np.zeros(arr.shape + (3,), dtype=interp_y_coords.dtype)
    interp_lch[..., 0] = np.interp(arr, interp_xp_coords, interp_y_coords[..., 0])
    interp_lch[..., 1] = np.interp(arr, interp_xp_coords, interp_y_coords[..., 1])
    interp_y_coords[..., 2] = np.unwrap(interp_y_coords[..., 2])
    interp_lch[..., 2] = np.interp(arr, interp_xp_coords, interp_y_coords[..., 2])
    interp_lch[..., 2] = _ununwrap(interp_lch[..., 2])
    new_rgb = lch2rgb(interp_lch)
    return [new_rgb[..., 0], new_rgb[..., 1], new_rgb[..., 2]]


def _ununwrap(input_radians):
    """Undo the operations performed by numpy unwrap.

    Taken from https://stackoverflow.com/a/15927914/433202

    """
    return (input_radians + np.pi) % (2 * np.pi) - np.pi


def _interpolate_alpha(arr, colors, values):
    alpha = [np.interp(arr,
                       np.array(values),
                       np.array(colors)[:, i + 3])
             for i in range(np.array(colors).shape[1] - 3)]
    return alpha


def _mask_channels(channels, arr):
    """Mask the channels if arr is a masked array."""
    return [_mask_array(channel, arr) for channel in channels]


def _mask_array(new_array, arr):
    """Mask new_array with the mask from array."""
    try:
        return np.ma.array(new_array, mask=arr.mask)
    except AttributeError:
        return new_array


def palettize(arr, colors, values):
    """Apply *colors* to *data* from start *values*.

    Args:
        arr (numpy array, numpy masked array, dask array):
            data to be palettized.
        colors (numpy array):
            the colors to use (R, G, B)
        values (numpy array):
            the values corresponding to the colors in the array
    """
    if can_be_block_mapped(arr):
        return _palettize_dask(arr, colors, values), tuple(colors)
    else:
        return _palettize(arr, values), tuple(colors)


def _palettize_dask(darr, colors, values):
    """Apply a palette to a dask array."""
    return darr.map_blocks(_palettize, values, dtype=int)


def _palettize(arr, values):
    """Apply palette to array."""
    new_arr = _digitize_array(arr, values)
    reshaped_array = new_arr.reshape(arr.shape)
    return _mask_array(reshaped_array, arr)


def _digitize_array(arr, values):
    if values[0] <= values[-1]:
        # monotonic increasing values
        outside_range_bin = max(np.nanmax(arr), values.max()) + np.int64(1)
        right = False
    else:
        # monotonic decreasing values
        outside_range_bin = min(np.nanmin(arr), values.min()) - np.int64(1)
        right = True
    bins = np.concatenate((values, [outside_range_bin]))

    new_arr = np.digitize(arr.ravel(), bins, right=right)
    new_arr -= 1
    new_arr = new_arr.clip(min=0, max=len(values) - 1)
    return new_arr


class Colormap(object):
    """The colormap object.

    Args:
        *args: Series of (value, color) tuples. These positional arguments
            are only used if the ``values`` and ``colors`` keyword arguments
            aren't provided.
        values: One dimensional array-like of control points where
            each corresponding color is applied. Must be the same number of
            elements as colors and must be monotonic.
        colors: Two dimensional array-like of RGB or RGBA colors where each
            color is applied to a specific control point. Must be the same
            number of colors as control points (values). Colors should be
            floating point numbers between 0 and 1.

    Initialize with tuples of (value, (colors)), like this::

      Colormap((-75.0, (1.0, 1.0, 0.0)),
               (-40.0001, (0.0, 1.0, 1.0)),
               (-40.0, (1, 1, 1)),
               (30.0, (0, 0, 0)))

    You can also concatenate colormaps together, try::

      cm = cm1 + cm2

    """

    def __init__(self, *tuples, **kwargs):
        """Set up the instance."""
        if 'colors' in kwargs and 'values' in kwargs:
            values = kwargs['values']
            colors = kwargs['colors']
        elif 'colors' in kwargs or 'values' in kwargs:
            raise ValueError("Both 'colors' and 'values' must be provided.")
        else:
            values = [a for (a, b) in tuples]
            colors = [b for (a, b) in tuples]
        self.values = np.array(values)
        self.colors = self._validate_colors(colors)
        if self.values.shape[0] != self.colors.shape[0]:
            raise ValueError("'values' and 'colors' should have the same "
                             "number of elements. Got "
                             f"{self.values.shape[0]} and {self.colors.shape[0]}.")

    def _validate_colors(self, colors):
        colors = np.asarray(colors)
        if colors.ndim != 2 or colors.shape[-1] not in (3, 4):
            raise ValueError("Colormap 'colors' must be RGB or RGBA. Got unexpected shape: {}".format(colors.shape))
        if not np.issubdtype(colors.dtype, np.floating):
            warnings.warn("Colormap 'colors' should be floating point numbers between 0 and 1.", stacklevel=3)
            colors = colors.astype(np.float64)
        return colors

    def colorize(self, data):
        """Colorize a monochromatic array *data*, based on the current colormap."""
        return colorize(data, self.colors, self.values)

    def palettize(self, data):
        """Palettize a monochromatic array *data* based on the current colormap."""
        return palettize(data, self.colors, self.values)

    def to_rgb(self):
        """Return colormap with RGB colors.

        If already RGB then the same instance is returned.
        If an Alpha channel exists in the colormap, it is dropped.

        """
        if self.colors.shape[-1] == 3:
            return self

        values = self.values.copy()
        colors = self.colors.copy()
        return Colormap(
            values=values,
            colors=colors[:, :3]
        )

    def to_rgba(self):
        """Return colormap with RGBA colors.

        If already RGBA then the same instance is returned.
        If not already RGBA, a completely opaque (1.0) color

        """
        if self.colors.shape[-1] == 4:
            return self

        values = self.values.copy()
        colors = np.empty((self.colors.shape[0], 4), dtype=self.colors.dtype)
        colors[:, :3] = self.colors
        colors[:, 3] = 1.0
        return Colormap(
            values=values,
            colors=colors
        )

    def __add__(self, other):
        """Append colormap together."""
        old, other = self._normalize_color_arrays(self, other)
        values = np.concatenate((old.values, other.values))
        if not self._monotonic_one_direction(values):
            raise ValueError("Merged colormap 'values' are not monotonically "
                             "increasing, monotonically decreasing, or equal.")
        colors = np.concatenate((old.colors, other.colors))
        return Colormap(
            values=values,
            colors=colors,
        )

    @staticmethod
    def _monotonic_one_direction(values):
        delta = np.diff(values)
        all_increasing = (delta >= 0).all()
        all_decreasing = (delta <= 0).all()
        return all_increasing or all_decreasing

    @staticmethod
    def _normalize_color_arrays(cmap1, cmap2):
        num_bands1 = cmap1.colors.shape[-1]
        num_bands2 = cmap2.colors.shape[-1]
        if num_bands1 == num_bands2:
            return cmap1, cmap2
        return cmap1.to_rgba(), cmap2.to_rgba()

    def reverse(self, inplace=True):
        """Reverse the current colormap in place.

        Args:
            inplace (bool): If True (default), modify the colors of this
                Colormap inplace. If False, return a new instance.

        """
        colors = np.flipud(self.colors)
        if not inplace:
            return Colormap(
                values=self.values.copy(),
                colors=colors
            )
        self.colors = colors
        return self

    def set_range(self, min_val, max_val, inplace=True):
        """Set the range of the colormap to [*min_val*, *max_val*].

        The Colormap's values will match the range specified even if "min_val"
        is greater than "max_val". To flip the order of the colors, use
        :meth:`reversed`.

        Args:
            min_val (float): New minimum value for the control points in
                this colormap.
            max_val (float): New maximum value for the control points in
                this colormap.
            inplace (bool): If True (default), modify the values inplace.
                If False, return a new Colormap instance.

        """
        cmap = self
        values = (((cmap.values * 1.0 - cmap.values[0]) /
                   (cmap.values[-1] - cmap.values[0]))
                  * (max_val - min_val) + min_val)
        if not inplace:
            return Colormap(
                values=values,
                colors=cmap.colors.copy()
            )

        cmap.values = values
        return cmap

    def set_alpha_range(self, min_alpha, max_alpha, inplace=True):
        """Set the colormap alpha channel between two values in linear steps.

        If the input colormap does not have an alpha channel,
        it will be added to it. If an alpha channel is already existing,
        the values will be overwritten.

        The min and max values shall be between 0 (completely transparent)
        and 1 (completely opaque).

        Args:
            min_alpha (float): Start value of the alpha channel [0-1]
            max_alpha (float): End value of the alpha channel [0-1]
            inplace (bool): If True (default), modify the values inplace.
                If False, return a new Colormap instance.

        """
        if inplace:
            cmap = self
        else:
            cmap = Colormap(
                values=self.values.copy(),
                colors=self.colors.copy())

        alpha = np.linspace(min_alpha,
                            max_alpha,
                            self.colors.shape[0])

        if cmap.colors.shape[1] == 4:
            cmap.colors[:, 3] = alpha
        else:
            cmap.colors = np.column_stack((cmap.colors, alpha))

        return cmap

    def to_rio(self):
        """Convert the colormap to a rasterio colormap.

        Note that rasterio requires color tables to have round integer value
        control points. This method assumes that the range of this Colormap
        is already in the desired output range and to avoid issues with
        rasterio will round the values and convert them to unsigned integers.
        """
        colors = (((self.colors * 1.0 - self.colors.min()) /
                   (self.colors.max() - self.colors.min())) * 255)
        # rasterio doesn't allow non-integer colormap values
        values = np.round(self.values).astype(np.uint)
        return dict(zip(values, tuple(map(tuple, colors))))

    def to_csv(
            self,
            filename: Optional[str] = None,
            color_scale: Number = 255,
    ) -> Optional[str]:
        """Save Colormap to a comma-separated text file or string.

        The CSV data will have 4 to 5 columns for each row where each
        each row will contain the value (V), red (R), green (B), blue (B),
        and if configured alpha (A).

        The values will remain in whatever range is currently set on the
        colormap. The colors of the colormap (assumed to be between 0 and 1)
        will be multiplied by 255 to produce a traditional unsigned 8-bit
        integer value.

        Args:
            filename: The filename of the CSV file to save to.
                If not provided or None a string is returned with the contents.
            color_scale: Scale colors by this factor before converting to a
                CSV. Colors are stored in the Colormap in a 0 to 1 range.
                Defaults to 255. If not equal to 1 values are converted to
                integers too.

        """
        with _file_or_stringio(filename) as csv_file:
            for value, color in zip(self.values, self.colors):
                scaled_color = [x * color_scale for x in color]
                if color_scale != 1.0:
                    scaled_color = [int(x) for x in scaled_color]
                csv_file.write(",".join(["{:0.6f}".format(value)] + [str(x) for x in scaled_color]) + "\n")
        if isinstance(csv_file, StringIO):
            return csv_file.getvalue()

    @classmethod
    def from_file(
            cls,
            filename: str,
            colormap_mode: Optional[str] = None,
            color_scale: Number = 255,
    ):
        """Create Colormap from a comma-separated or binary file of colormap data.

        Args:
            filename: Filename of a binary or CSV file
            colormap_mode: Force the scheme of the colormap data (ex. RGBA).
                See information below on other possible values and how they
                are interpreted. By default this is determined based on the
                number of columns in the data.
            color_scale: The maximum possible color value in the colormap data
                provided. For example, if the colors in the provided data were
                8-bit unsigned integers this should be 255 (the default). This
                value will be used to normalize the colors from 0 to 1.

        Colormaps can be loaded from ``.npy``, ``.npz``, or comma-separated text
        files. Numpy (npy/npz) files should be 2D arrays with rows for each color.
        Comma-separated files should have a row for each color with each column
        representing a single value/channel. A filename
        ending with ``.npy`` or ``.npz`` is read as a numpy file with
        :func:`numpy.load`. All other extensions are
        read as a comma-separated file. For ``.npz`` files the data must be stored
        as a positional list where the first element represents the colormap to
        use. See :func:`numpy.savez` for more information.

        The colormap is interpreted as 1 of 4 different "colormap modes":
        ``RGB``, ``RGBA``, ``VRGB``, or ``VRGBA``. The
        colormap mode can be forced with the ``colormap_mode`` keyword
        argument. If it is not provided then a default will be chosen
        based on the number of columns in the array (3: RGB, 4: VRGB, 5: VRGBA).

        The "V" in the possible colormap modes represents the control value of
        where that color should be applied. If "V" is not provided in the colormap
        data it defaults to the row index in the colormap array (0, 1, 2, ...)
        divided by the total number of colors to produce a number between 0 and 1.
        See the "Set Range" section below for more information.
        The remaining elements in the colormap array represent the Red (R),
        Green (G), and Blue (B) color to be mapped to.

        See the "Color Scale" section below for more information on the value
        range of provided numbers.

        To read from a string containing CSV, use :meth:`~Colormap.from_csv`.

        To get a named colormap, use :meth:`~Colormap.from_name` or load the
        colormap directly as a module attribute.

        To get a colormap from an ndarray, use :meth:`~Colormap.from_ndarray`.

        **Color Scale**

        By default colors are expected to be in a 0-255 range. This
        can be overridden by specifying ``color_scale`` keyword argument.
        A common alternative to 255 is ``1`` to specify floating
        point numbers between 0 and 1. The resulting Colormap uses the normalized
        color values (0-1).
        """
        if _is_actually_a_csv_string(filename):
            warnings.warn(
                "Passing a data string to Colormap.from_file is deprecated. "
                "Please use Colormap.from_string.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return cls.from_string(filename, colormap_mode, color_scale)
        values, colors = _get_values_colors_from_file(filename, colormap_mode, color_scale)
        return cls(values=values, colors=colors)

    @classmethod
    def from_string(cls, string, *args, **kwargs):
        """Create colormap from string.

        Create a colormap from a string that contains comma seperated values
        (CSV).

        To read from an external file that contains CSV, use :meth:`from_csv`.

        Args:
            string (str): String containing CSV.  Must have no less than three
                and no more than five columns and describe entirely numeric
                data.
            colormap_mode (str or None): Optional. Can be None, "RGB", "RGBA", "VRGB", or
                "VRGBA".  If None (default), this is inferred from the dimensions of
                the data contained in the CSV.  Modes starting with V have in
                the first column the values to which the color relates.
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
        """
        openfile = StringIO(string)
        cmap_data = _read_colormap_data_from_file(openfile)
        return cls.from_ndarray(cmap_data, *args, **kwargs)

    @classmethod
    def from_np(cls, path, *args, **kwargs):
        """Create Colormap from a numpy-file.

        Create a colormap from a numpy data file ``.npy`` or ``.npz``.

        The data should contain at least three and at most five columns.

        Args:
            path (str or Pathlib.Path): Path to file containing numpy data.
            colormap_mode (str or None): Optional. Can be None, "RGB", "RGBA", "VRGB", or
                "VRGBA".  If None (default), this is inferred from the dimensions of
                the data contained in the CSV.  Modes starting with V have in
                the first column the values to which the color relates.
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
        """
        cmap_data = _read_colormap_data_from_np(path)
        return cls.from_ndarray(cmap_data, *args, **kwargs)

    @classmethod
    def from_csv(cls, path, colormap_mode=None, color_scale=255):
        """Create Colormap from CSV file.

        Create a Colormap from a file that contains comma seperated values
        (CSV).

        To read from a string that contains CSV, use :meth:`from_string`.

        Args:
            string (str or pathlib.Path): Path to file containing CSV.
                The CSV must have at least three and at most five columns and
                describe entirely numeric data.
            colormap_mode (str or None): Optional. Can be None, "RGB", "RGBA", "VRGB", or
                "VRGBA".  If None (default), this is inferred from the dimensions of
                the data contained in the CSV.  Modes starting with V have in
                the first column the values to which the color relates.
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
        """
        cmap_data = np.loadtxt(path, delimiter=",")
        return cls.from_ndarray(cmap_data, colormap_mode, color_scale)

    @classmethod
    def from_ndarray(cls, cmap_data, colormap_mode=None, color_scale=255):
        """Create Colormap from ndarray.

        Create a colormap from a numpy data array.

        The data should contain at least three and at most five columns.

        For historical reasons, this method exists alongside
        :meth:`from_sequence_of_colors` and :meth:`from_array_with_metadata` despite similar
        functionality.

        Args:
            cmap_data (ndarray): Array describing the colours.
                Must have at least three and at most five columns and
                have a numeric dtype.
            colormap_mode (str or None): Optional. Can be None, "RGB", "RGBA", "VRGB", or
                "VRGBA".  If None (default), this is inferred from the dimensions of
                the data contained in the CSV.  Modes starting with V have in
                the first column the values to which the color relates.
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
        """
        values, colors = _get_values_colors_from_ndarray(cmap_data, colormap_mode, color_scale)
        return cls(values=values, colors=colors)

    @classmethod
    def from_name(cls, name):
        """Return named colormap.

        Return a colormap by name.  Supported colormaps are the ones defined in
        the module namespace.

        Args:
            name (str): Name of colormap.
        """
        cmap = getattr(sys.modules[__name__], name)
        return copy.copy(cmap)

    @classmethod
    def from_sequence_of_colors(cls, colors, values=None, color_scale=255):
        """Create Colormap from sequence of colors.

        Create a colormap from a sequence of colors, such as a list of colors.
        If values is not given, assume values between 0 and 1, linearly spaced
        according to the total number of colors.

        For historical reasons, this method exists alongside
        :meth:`from_ndarray` and :meth:`from_array_with_metadata` despite similar
        functionality.

        Args:
            colors (Sequence): List of colors, where each element must itself
                be a sequence of 3 or 4 numbers (RGB or RGBA).
            values (array, optional): Values associated with the colors.  If
                not given, assume linear between 0 and 1.
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
        """
        # this method was moved from satpy. where it was in
        # satpy.enhancements.create_colormap
        # then it was refactored/rewritten
        color_array = np.array(colors)
        if values is None:
            values = np.linspace(0, 1, len(colors))
        else:
            values = np.asarray(values)
        color_array = np.concatenate((values[:, np.newaxis], color_array), axis=1)
        return cls.from_ndarray(
            color_array,
            "VRGB" if color_array.shape[1] == 4 else "VRGBA",
            color_scale=color_scale)

    @classmethod
    def from_array_with_metadata(
            cls, palette, dtype, color_scale=255,
            valid_range=None, scale_factor=1, add_offset=0,
            remove_last=True):
        """Create Colormap from an array with metadata.

        Create a colormap from an array with associated metadata, either in
        attributes or passed on to the function.

        For historical reasons, this method exists alongside
        :meth:`from_ndarray` and :meth:`from_sequence_of_colors` despite similar
        functionality.

        If ``palette`` is an xarray dataarray with the attribute
        ``palette_meanings``, those meanings are interpreted as values
        associated with the colormap.

        If no values can be interpreted from the metadata, values will be
        linearly interpolated between 0 and 255 (if ``dtype`` is ``np.uint8``)
        or according to ``valid_range``.

        Args:
            palette (ndarray or xarray.DataArray)
                Array describing colors, possibly with metadata.  If it has a
                ``palette_meanings`` attribute, this will be used for color
                interpretation.
            dtype
                dtype for the colormap
            color_scale (number): The value that represents white in the
                numbers describing the colors. Defaults to 255, could also be 1
                or something else.
            valid_range
                valid range for colors, if colormap is not of dtype uint8
            scale_factor
                scale factor to apply to the colormap
            add_offset
                add offset to apply to the colormap
            remove_last
                Remove the last value if the array has no metadata associated.
                Defaults to true for historical reasons.

        """
        # this method was moved from satpy, where it was in
        # satpy.composites.ColormapCompositor.build_colormap
        #
        # then it was refactored/rewritten for trollimage
        squeezed_palette = np.asanyarray(palette).squeeze()
        set_range = True
        if hasattr(palette, 'attrs') and 'palette_meanings' in palette.attrs:
            set_range = False
            values = np.asarray(palette.attrs['palette_meanings'])
        else:
            # remove the last value because monkeys don't like water sprays
            # on a more serious note, I don't know why we are removing the last
            # value here, but this behaviour was copied from ancient satpy code
            values = np.arange(squeezed_palette.shape[0] - remove_last)
            if remove_last:
                squeezed_palette = squeezed_palette[:-remove_last, :]

        color_array = np.concatenate((values[:, np.newaxis], squeezed_palette), axis=1)
        colormap = cls.from_ndarray(
            color_array,
            "VRGB" if color_array.shape[1] == 4 else "VRGBA",
            color_scale=color_scale)
        if dtype == np.dtype("uint8"):
            return colormap

        if valid_range is not None:
            if set_range:
                colormap.set_range(
                    *(np.array(valid_range) * scale_factor
                      + add_offset))
            return colormap

        raise AttributeError("Data need to have either a valid_range or be of type uint8" +
                             " in order to be displayable with an attached color-palette!")


def _is_actually_a_csv_string(string):
    """Try to guess whether this string contains CSV."""
    return string.count("\n") > 0 and string.count(",") > 0


def _get_values_colors_from_file(filename, colormap_mode, color_scale):
    data = _read_colormap_data_from_file(filename)
    return _get_values_colors_from_ndarray(data, colormap_mode, color_scale)


def _get_values_colors_from_ndarray(data, colormap_mode, color_scale):
    cols = data.shape[1]
    default_modes = {
        3: 'RGB',
        4: 'VRGB',
        5: 'VRGBA'
    }
    default_mode = default_modes.get(cols)
    if colormap_mode is None:
        colormap_mode = default_mode
    if colormap_mode is None or len(colormap_mode) != cols:
        raise ValueError(
            "Unexpected colormap shape for mode '{}'".format(colormap_mode))
    rows = data.shape[0]
    if colormap_mode[0] == 'V':
        colors = data[:, 1:]
        if color_scale != 1:
            colors = data[:, 1:] / float(color_scale)
        values = data[:, 0]
    else:
        colors = data
        if color_scale != 1:
            colors = colors / float(color_scale)
        values = np.arange(rows) / float(rows - 1)
    return values, colors


def _read_colormap_data_from_file(filename_or_file_obj):
    if isinstance(filename_or_file_obj, str):
        ext = os.path.splitext(filename_or_file_obj)[1]
        if ext in (".npy", ".npz"):
            return _read_colormap_data_from_np(filename_or_file_obj)
    # CSV file or file-like object of CSV string data
    return np.loadtxt(filename_or_file_obj, delimiter=",")


def _read_colormap_data_from_np(path):
    path = pathlib.Path(path)
    file_content = np.load(path)
    if path.suffix == ".npz":
        # .npz is a collection
        # assume position list-like and get the first element
        file_content = file_content["arr_0"]
    return file_content


# matlab jet "#00007F", "blue", "#007FFF", "cyan", "#7FFF7F", "yellow",
# "#FF7F00", "red", "#7F0000"

rainbow = Colormap((0.000, (0.0, 0.0, 0.5)),
                   (0.125, (0.0, 0.0, 1.0)),
                   (0.250, (0.0, 0.5, 1.0)),
                   (0.375, (0.0, 1.0, 1.0)),
                   (0.500, (0.5, 1.0, 0.5)),
                   (0.625, (1.0, 1.0, 0.0)),
                   (0.750, (1.0, 0.5, 0.0)),
                   (0.875, (1.0, 0.0, 0.0)),
                   (1.000, (0.5, 0.0, 0.0)))

# * Colors from www.ColorBrewer.org by Cynthia A. Brewer, Geography,
# * Pennsylvania State University.

# * Sequential Colormaps *

blues = Colormap(
    (0 / 8, (247 / 255, 251 / 255, 255 / 255)),
    (1 / 8, (222 / 255, 235 / 255, 247 / 255)),
    (2 / 8, (198 / 255, 219 / 255, 239 / 255)),
    (3 / 8, (158 / 255, 202 / 255, 225 / 255)),
    (4 / 8, (107 / 255, 174 / 255, 214 / 255)),
    (5 / 8, (66 / 255, 146 / 255, 198 / 255)),
    (6 / 8, (33 / 255, 113 / 255, 181 / 255)),
    (7 / 8, (8 / 255, 81 / 255, 156 / 255)),
    (8 / 8, (8 / 255, 48 / 255, 107 / 255)),
)

bugn = Colormap(
    (0 / 8, (247 / 255, 252 / 255, 253 / 255)),
    (1 / 8, (229 / 255, 245 / 255, 249 / 255)),
    (2 / 8, (204 / 255, 236 / 255, 230 / 255)),
    (3 / 8, (153 / 255, 216 / 255, 201 / 255)),
    (4 / 8, (102 / 255, 194 / 255, 164 / 255)),
    (5 / 8, (65 / 255, 174 / 255, 118 / 255)),
    (6 / 8, (35 / 255, 139 / 255, 69 / 255)),
    (7 / 8, (0 / 255, 109 / 255, 44 / 255)),
    (8 / 8, (0 / 255, 68 / 255, 27 / 255)),
)

bupu = Colormap(
    (0 / 8, (247 / 255, 252 / 255, 253 / 255)),
    (1 / 8, (224 / 255, 236 / 255, 244 / 255)),
    (2 / 8, (191 / 255, 211 / 255, 230 / 255)),
    (3 / 8, (158 / 255, 188 / 255, 218 / 255)),
    (4 / 8, (140 / 255, 150 / 255, 198 / 255)),
    (5 / 8, (140 / 255, 107 / 255, 177 / 255)),
    (6 / 8, (136 / 255, 65 / 255, 157 / 255)),
    (7 / 8, (129 / 255, 15 / 255, 124 / 255)),
    (8 / 8, (77 / 255, 0 / 255, 75 / 255)),
)

gnbu = Colormap(
    (0 / 8, (247 / 255, 252 / 255, 240 / 255)),
    (1 / 8, (224 / 255, 243 / 255, 219 / 255)),
    (2 / 8, (204 / 255, 235 / 255, 197 / 255)),
    (3 / 8, (168 / 255, 221 / 255, 181 / 255)),
    (4 / 8, (123 / 255, 204 / 255, 196 / 255)),
    (5 / 8, (78 / 255, 179 / 255, 211 / 255)),
    (6 / 8, (43 / 255, 140 / 255, 190 / 255)),
    (7 / 8, (8 / 255, 104 / 255, 172 / 255)),
    (8 / 8, (8 / 255, 64 / 255, 129 / 255)),
)

greens = Colormap(
    (0 / 8, (247 / 255, 252 / 255, 245 / 255)),
    (1 / 8, (229 / 255, 245 / 255, 224 / 255)),
    (2 / 8, (199 / 255, 233 / 255, 192 / 255)),
    (3 / 8, (161 / 255, 217 / 255, 155 / 255)),
    (4 / 8, (116 / 255, 196 / 255, 118 / 255)),
    (5 / 8, (65 / 255, 171 / 255, 93 / 255)),
    (6 / 8, (35 / 255, 139 / 255, 69 / 255)),
    (7 / 8, (0 / 255, 109 / 255, 44 / 255)),
    (8 / 8, (0 / 255, 68 / 255, 27 / 255)),
)

greys = Colormap(
    (0 / 8, (255 / 255, 255 / 255, 255 / 255)),
    (1 / 8, (240 / 255, 240 / 255, 240 / 255)),
    (2 / 8, (217 / 255, 217 / 255, 217 / 255)),
    (3 / 8, (189 / 255, 189 / 255, 189 / 255)),
    (4 / 8, (150 / 255, 150 / 255, 150 / 255)),
    (5 / 8, (115 / 255, 115 / 255, 115 / 255)),
    (6 / 8, (82 / 255, 82 / 255, 82 / 255)),
    (7 / 8, (37 / 255, 37 / 255, 37 / 255)),
    (8 / 8, (0 / 255, 0 / 255, 0 / 255)),
)

oranges = Colormap(
    (0 / 8, (255 / 255, 245 / 255, 235 / 255)),
    (1 / 8, (254 / 255, 230 / 255, 206 / 255)),
    (2 / 8, (253 / 255, 208 / 255, 162 / 255)),
    (3 / 8, (253 / 255, 174 / 255, 107 / 255)),
    (4 / 8, (253 / 255, 141 / 255, 60 / 255)),
    (5 / 8, (241 / 255, 105 / 255, 19 / 255)),
    (6 / 8, (217 / 255, 72 / 255, 1 / 255)),
    (7 / 8, (166 / 255, 54 / 255, 3 / 255)),
    (8 / 8, (127 / 255, 39 / 255, 4 / 255)),
)

orrd = Colormap(
    (0 / 8, (255 / 255, 247 / 255, 236 / 255)),
    (1 / 8, (254 / 255, 232 / 255, 200 / 255)),
    (2 / 8, (253 / 255, 212 / 255, 158 / 255)),
    (3 / 8, (253 / 255, 187 / 255, 132 / 255)),
    (4 / 8, (252 / 255, 141 / 255, 89 / 255)),
    (5 / 8, (239 / 255, 101 / 255, 72 / 255)),
    (6 / 8, (215 / 255, 48 / 255, 31 / 255)),
    (7 / 8, (179 / 255, 0 / 255, 0 / 255)),
    (8 / 8, (127 / 255, 0 / 255, 0 / 255)),
)

pubu = Colormap(
    (0 / 8, (255 / 255, 247 / 255, 251 / 255)),
    (1 / 8, (236 / 255, 231 / 255, 242 / 255)),
    (2 / 8, (208 / 255, 209 / 255, 230 / 255)),
    (3 / 8, (166 / 255, 189 / 255, 219 / 255)),
    (4 / 8, (116 / 255, 169 / 255, 207 / 255)),
    (5 / 8, (54 / 255, 144 / 255, 192 / 255)),
    (6 / 8, (5 / 255, 112 / 255, 176 / 255)),
    (7 / 8, (4 / 255, 90 / 255, 141 / 255)),
    (8 / 8, (2 / 255, 56 / 255, 88 / 255)),
)

pubugn = Colormap(
    (0 / 8, (255 / 255, 247 / 255, 251 / 255)),
    (1 / 8, (236 / 255, 226 / 255, 240 / 255)),
    (2 / 8, (208 / 255, 209 / 255, 230 / 255)),
    (3 / 8, (166 / 255, 189 / 255, 219 / 255)),
    (4 / 8, (103 / 255, 169 / 255, 207 / 255)),
    (5 / 8, (54 / 255, 144 / 255, 192 / 255)),
    (6 / 8, (2 / 255, 129 / 255, 138 / 255)),
    (7 / 8, (1 / 255, 108 / 255, 89 / 255)),
    (8 / 8, (1 / 255, 70 / 255, 54 / 255)),
)

purd = Colormap(
    (0 / 8, (247 / 255, 244 / 255, 249 / 255)),
    (1 / 8, (231 / 255, 225 / 255, 239 / 255)),
    (2 / 8, (212 / 255, 185 / 255, 218 / 255)),
    (3 / 8, (201 / 255, 148 / 255, 199 / 255)),
    (4 / 8, (223 / 255, 101 / 255, 176 / 255)),
    (5 / 8, (231 / 255, 41 / 255, 138 / 255)),
    (6 / 8, (206 / 255, 18 / 255, 86 / 255)),
    (7 / 8, (152 / 255, 0 / 255, 67 / 255)),
    (8 / 8, (103 / 255, 0 / 255, 31 / 255)),
)

purples = Colormap(
    (0 / 8, (252 / 255, 251 / 255, 253 / 255)),
    (1 / 8, (239 / 255, 237 / 255, 245 / 255)),
    (2 / 8, (218 / 255, 218 / 255, 235 / 255)),
    (3 / 8, (188 / 255, 189 / 255, 220 / 255)),
    (4 / 8, (158 / 255, 154 / 255, 200 / 255)),
    (5 / 8, (128 / 255, 125 / 255, 186 / 255)),
    (6 / 8, (106 / 255, 81 / 255, 163 / 255)),
    (7 / 8, (84 / 255, 39 / 255, 143 / 255)),
    (8 / 8, (63 / 255, 0 / 255, 125 / 255)),
)

rdpu = Colormap(
    (0 / 8, (255 / 255, 247 / 255, 243 / 255)),
    (1 / 8, (253 / 255, 224 / 255, 221 / 255)),
    (2 / 8, (252 / 255, 197 / 255, 192 / 255)),
    (3 / 8, (250 / 255, 159 / 255, 181 / 255)),
    (4 / 8, (247 / 255, 104 / 255, 161 / 255)),
    (5 / 8, (221 / 255, 52 / 255, 151 / 255)),
    (6 / 8, (174 / 255, 1 / 255, 126 / 255)),
    (7 / 8, (122 / 255, 1 / 255, 119 / 255)),
    (8 / 8, (73 / 255, 0 / 255, 106 / 255)),
)

reds = Colormap(
    (0 / 8, (255 / 255, 245 / 255, 240 / 255)),
    (1 / 8, (254 / 255, 224 / 255, 210 / 255)),
    (2 / 8, (252 / 255, 187 / 255, 161 / 255)),
    (3 / 8, (252 / 255, 146 / 255, 114 / 255)),
    (4 / 8, (251 / 255, 106 / 255, 74 / 255)),
    (5 / 8, (239 / 255, 59 / 255, 44 / 255)),
    (6 / 8, (203 / 255, 24 / 255, 29 / 255)),
    (7 / 8, (165 / 255, 15 / 255, 21 / 255)),
    (8 / 8, (103 / 255, 0 / 255, 13 / 255)),
)

ylgn = Colormap(
    (0 / 8, (255 / 255, 255 / 255, 229 / 255)),
    (1 / 8, (247 / 255, 252 / 255, 185 / 255)),
    (2 / 8, (217 / 255, 240 / 255, 163 / 255)),
    (3 / 8, (173 / 255, 221 / 255, 142 / 255)),
    (4 / 8, (120 / 255, 198 / 255, 121 / 255)),
    (5 / 8, (65 / 255, 171 / 255, 93 / 255)),
    (6 / 8, (35 / 255, 132 / 255, 67 / 255)),
    (7 / 8, (0 / 255, 104 / 255, 55 / 255)),
    (8 / 8, (0 / 255, 69 / 255, 41 / 255)),
)

ylgnbu = Colormap(
    (0 / 8, (255 / 255, 255 / 255, 217 / 255)),
    (1 / 8, (237 / 255, 248 / 255, 177 / 255)),
    (2 / 8, (199 / 255, 233 / 255, 180 / 255)),
    (3 / 8, (127 / 255, 205 / 255, 187 / 255)),
    (4 / 8, (65 / 255, 182 / 255, 196 / 255)),
    (5 / 8, (29 / 255, 145 / 255, 192 / 255)),
    (6 / 8, (34 / 255, 94 / 255, 168 / 255)),
    (7 / 8, (37 / 255, 52 / 255, 148 / 255)),
    (8 / 8, (8 / 255, 29 / 255, 88 / 255)),
)

ylorbr = Colormap(
    (0 / 8, (255 / 255, 255 / 255, 229 / 255)),
    (1 / 8, (255 / 255, 247 / 255, 188 / 255)),
    (2 / 8, (254 / 255, 227 / 255, 145 / 255)),
    (3 / 8, (254 / 255, 196 / 255, 79 / 255)),
    (4 / 8, (254 / 255, 153 / 255, 41 / 255)),
    (5 / 8, (236 / 255, 112 / 255, 20 / 255)),
    (6 / 8, (204 / 255, 76 / 255, 2 / 255)),
    (7 / 8, (153 / 255, 52 / 255, 4 / 255)),
    (8 / 8, (102 / 255, 37 / 255, 6 / 255)),
)

ylorrd = Colormap(
    (0 / 7, (255 / 255, 255 / 255, 204 / 255)),
    (1 / 7, (255 / 255, 237 / 255, 160 / 255)),
    (2 / 7, (254 / 255, 217 / 255, 118 / 255)),
    (3 / 7, (254 / 255, 178 / 255, 76 / 255)),
    (4 / 7, (253 / 255, 141 / 255, 60 / 255)),
    (5 / 7, (252 / 255, 78 / 255, 42 / 255)),
    (6 / 7, (227 / 255, 26 / 255, 28 / 255)),
    (7 / 7, (177 / 255, 0 / 255, 38 / 255)),
)

sequential_colormaps = {
    "blues": blues,
    "bugn": bugn,
    "bupu": bupu,
    "gnbu": gnbu,
    "greens": greens,
    "greys": greys,
    "oranges": oranges,
    "orrd": orrd,
    "pubu": pubu,
    "pubugn": pubugn,
    "purd": purd,
    "purples": purples,
    "rdpu": rdpu,
    "reds": reds,
    "ylgn": ylgn,
    "ylgnbu": ylgnbu,
    "ylorbr": ylorbr,
    "ylorrd": ylorrd,
}

# * Diverging Colormaps *

brbg = Colormap(
    (0 / 10, (84 / 255, 48 / 255, 5 / 255)),
    (1 / 10, (140 / 255, 81 / 255, 10 / 255)),
    (2 / 10, (191 / 255, 129 / 255, 45 / 255)),
    (3 / 10, (223 / 255, 194 / 255, 125 / 255)),
    (4 / 10, (246 / 255, 232 / 255, 195 / 255)),
    (5 / 10, (245 / 255, 245 / 255, 245 / 255)),
    (6 / 10, (199 / 255, 234 / 255, 229 / 255)),
    (7 / 10, (128 / 255, 205 / 255, 193 / 255)),
    (8 / 10, (53 / 255, 151 / 255, 143 / 255)),
    (9 / 10, (1 / 255, 102 / 255, 94 / 255)),
    (10 / 10, (0 / 255, 60 / 255, 48 / 255)),
)

piyg = Colormap(
    (0 / 10, (142 / 255, 1 / 255, 82 / 255)),
    (1 / 10, (197 / 255, 27 / 255, 125 / 255)),
    (2 / 10, (222 / 255, 119 / 255, 174 / 255)),
    (3 / 10, (241 / 255, 182 / 255, 218 / 255)),
    (4 / 10, (253 / 255, 224 / 255, 239 / 255)),
    (5 / 10, (247 / 255, 247 / 255, 247 / 255)),
    (6 / 10, (230 / 255, 245 / 255, 208 / 255)),
    (7 / 10, (184 / 255, 225 / 255, 134 / 255)),
    (8 / 10, (127 / 255, 188 / 255, 65 / 255)),
    (9 / 10, (77 / 255, 146 / 255, 33 / 255)),
    (10 / 10, (39 / 255, 100 / 255, 25 / 255)),
)

prgn = Colormap(
    (0 / 10, (64 / 255, 0 / 255, 75 / 255)),
    (1 / 10, (118 / 255, 42 / 255, 131 / 255)),
    (2 / 10, (153 / 255, 112 / 255, 171 / 255)),
    (3 / 10, (194 / 255, 165 / 255, 207 / 255)),
    (4 / 10, (231 / 255, 212 / 255, 232 / 255)),
    (5 / 10, (247 / 255, 247 / 255, 247 / 255)),
    (6 / 10, (217 / 255, 240 / 255, 211 / 255)),
    (7 / 10, (166 / 255, 219 / 255, 160 / 255)),
    (8 / 10, (90 / 255, 174 / 255, 97 / 255)),
    (9 / 10, (27 / 255, 120 / 255, 55 / 255)),
    (10 / 10, (0 / 255, 68 / 255, 27 / 255)),
)

puor = Colormap(
    (0 / 10, (127 / 255, 59 / 255, 8 / 255)),
    (1 / 10, (179 / 255, 88 / 255, 6 / 255)),
    (2 / 10, (224 / 255, 130 / 255, 20 / 255)),
    (3 / 10, (253 / 255, 184 / 255, 99 / 255)),
    (4 / 10, (254 / 255, 224 / 255, 182 / 255)),
    (5 / 10, (247 / 255, 247 / 255, 247 / 255)),
    (6 / 10, (216 / 255, 218 / 255, 235 / 255)),
    (7 / 10, (178 / 255, 171 / 255, 210 / 255)),
    (8 / 10, (128 / 255, 115 / 255, 172 / 255)),
    (9 / 10, (84 / 255, 39 / 255, 136 / 255)),
    (10 / 10, (45 / 255, 0 / 255, 75 / 255)),
)

rdbu = Colormap(
    (0 / 10, (103 / 255, 0 / 255, 31 / 255)),
    (1 / 10, (178 / 255, 24 / 255, 43 / 255)),
    (2 / 10, (214 / 255, 96 / 255, 77 / 255)),
    (3 / 10, (244 / 255, 165 / 255, 130 / 255)),
    (4 / 10, (253 / 255, 219 / 255, 199 / 255)),
    (5 / 10, (247 / 255, 247 / 255, 247 / 255)),
    (6 / 10, (209 / 255, 229 / 255, 240 / 255)),
    (7 / 10, (146 / 255, 197 / 255, 222 / 255)),
    (8 / 10, (67 / 255, 147 / 255, 195 / 255)),
    (9 / 10, (33 / 255, 102 / 255, 172 / 255)),
    (10 / 10, (5 / 255, 48 / 255, 97 / 255)),
)

rdgy = Colormap(
    (0 / 10, (103 / 255, 0 / 255, 31 / 255)),
    (1 / 10, (178 / 255, 24 / 255, 43 / 255)),
    (2 / 10, (214 / 255, 96 / 255, 77 / 255)),
    (3 / 10, (244 / 255, 165 / 255, 130 / 255)),
    (4 / 10, (253 / 255, 219 / 255, 199 / 255)),
    (5 / 10, (255 / 255, 255 / 255, 255 / 255)),
    (6 / 10, (224 / 255, 224 / 255, 224 / 255)),
    (7 / 10, (186 / 255, 186 / 255, 186 / 255)),
    (8 / 10, (135 / 255, 135 / 255, 135 / 255)),
    (9 / 10, (77 / 255, 77 / 255, 77 / 255)),
    (10 / 10, (26 / 255, 26 / 255, 26 / 255)),
)

rdylbu = Colormap(
    (0 / 10, (165 / 255, 0 / 255, 38 / 255)),
    (1 / 10, (215 / 255, 48 / 255, 39 / 255)),
    (2 / 10, (244 / 255, 109 / 255, 67 / 255)),
    (3 / 10, (253 / 255, 174 / 255, 97 / 255)),
    (4 / 10, (254 / 255, 224 / 255, 144 / 255)),
    (5 / 10, (255 / 255, 255 / 255, 191 / 255)),
    (6 / 10, (224 / 255, 243 / 255, 248 / 255)),
    (7 / 10, (171 / 255, 217 / 255, 233 / 255)),
    (8 / 10, (116 / 255, 173 / 255, 209 / 255)),
    (9 / 10, (69 / 255, 117 / 255, 180 / 255)),
    (10 / 10, (49 / 255, 54 / 255, 149 / 255)),
)

rdylgn = Colormap(
    (0 / 10, (165 / 255, 0 / 255, 38 / 255)),
    (1 / 10, (215 / 255, 48 / 255, 39 / 255)),
    (2 / 10, (244 / 255, 109 / 255, 67 / 255)),
    (3 / 10, (253 / 255, 174 / 255, 97 / 255)),
    (4 / 10, (254 / 255, 224 / 255, 139 / 255)),
    (5 / 10, (255 / 255, 255 / 255, 191 / 255)),
    (6 / 10, (217 / 255, 239 / 255, 139 / 255)),
    (7 / 10, (166 / 255, 217 / 255, 106 / 255)),
    (8 / 10, (102 / 255, 189 / 255, 99 / 255)),
    (9 / 10, (26 / 255, 152 / 255, 80 / 255)),
    (10 / 10, (0 / 255, 104 / 255, 55 / 255)),
)

spectral = Colormap(
    (0 / 10, (158 / 255, 1 / 255, 66 / 255)),
    (1 / 10, (213 / 255, 62 / 255, 79 / 255)),
    (2 / 10, (244 / 255, 109 / 255, 67 / 255)),
    (3 / 10, (253 / 255, 174 / 255, 97 / 255)),
    (4 / 10, (254 / 255, 224 / 255, 139 / 255)),
    (5 / 10, (255 / 255, 255 / 255, 191 / 255)),
    (6 / 10, (230 / 255, 245 / 255, 152 / 255)),
    (7 / 10, (171 / 255, 221 / 255, 164 / 255)),
    (8 / 10, (102 / 255, 194 / 255, 165 / 255)),
    (9 / 10, (50 / 255, 136 / 255, 189 / 255)),
    (10 / 10, (94 / 255, 79 / 255, 162 / 255)),
)

diverging_colormaps = {
    "brbg": brbg,
    "piyg": piyg,
    "prgn": prgn,
    "puor": puor,
    "rdbu": rdbu,
    "rdgy": rdgy,
    "rdylbu": rdylbu,
    "rdylgn": rdylgn,
    "spectral": spectral,
}

# * Qualitative Colormaps *

accent = Colormap(
    (0, (127 / 255, 201 / 255, 127 / 255)),
    (1, (190 / 255, 174 / 255, 212 / 255)),
    (2, (253 / 255, 192 / 255, 134 / 255)),
    (3, (255 / 255, 255 / 255, 153 / 255)),
    (4, (56 / 255, 108 / 255, 176 / 255)),
    (5, (240 / 255, 2 / 255, 127 / 255)),
    (6, (191 / 255, 91 / 255, 23 / 255)),
    (7, (102 / 255, 102 / 255, 102 / 255)),
)

dark2 = Colormap(
    (0, (27 / 255, 158 / 255, 119 / 255)),
    (1, (217 / 255, 95 / 255, 2 / 255)),
    (2, (117 / 255, 112 / 255, 179 / 255)),
    (3, (231 / 255, 41 / 255, 138 / 255)),
    (4, (102 / 255, 166 / 255, 30 / 255)),
    (5, (230 / 255, 171 / 255, 2 / 255)),
    (6, (166 / 255, 118 / 255, 29 / 255)),
    (7, (102 / 255, 102 / 255, 102 / 255)),
)

paired = Colormap(
    (0, (166 / 255, 206 / 255, 227 / 255)),
    (1, (31 / 255, 120 / 255, 180 / 255)),
    (2, (178 / 255, 223 / 255, 138 / 255)),
    (3, (51 / 255, 160 / 255, 44 / 255)),
    (4, (251 / 255, 154 / 255, 153 / 255)),
    (5, (227 / 255, 26 / 255, 28 / 255)),
    (6, (253 / 255, 191 / 255, 111 / 255)),
    (7, (255 / 255, 127 / 255, 0 / 255)),
    (8, (202 / 255, 178 / 255, 214 / 255)),
    (9, (106 / 255, 61 / 255, 154 / 255)),
    (10, (255 / 255, 255 / 255, 153 / 255)),
    (11, (177 / 255, 89 / 255, 40 / 255)),
)

pastel1 = Colormap(
    (0, (251 / 255, 180 / 255, 174 / 255)),
    (1, (179 / 255, 205 / 255, 227 / 255)),
    (2, (204 / 255, 235 / 255, 197 / 255)),
    (3, (222 / 255, 203 / 255, 228 / 255)),
    (4, (254 / 255, 217 / 255, 166 / 255)),
    (5, (255 / 255, 255 / 255, 204 / 255)),
    (6, (229 / 255, 216 / 255, 189 / 255)),
    (7, (253 / 255, 218 / 255, 236 / 255)),
    (8, (242 / 255, 242 / 255, 242 / 255)),
)

pastel2 = Colormap(
    (0, (179 / 255, 226 / 255, 205 / 255)),
    (1, (253 / 255, 205 / 255, 172 / 255)),
    (2, (203 / 255, 213 / 255, 232 / 255)),
    (3, (244 / 255, 202 / 255, 228 / 255)),
    (4, (230 / 255, 245 / 255, 201 / 255)),
    (5, (255 / 255, 242 / 255, 174 / 255)),
    (6, (241 / 255, 226 / 255, 204 / 255)),
    (7, (204 / 255, 204 / 255, 204 / 255)),
)

set1 = Colormap(
    (0, (228 / 255, 26 / 255, 28 / 255)),
    (1, (55 / 255, 126 / 255, 184 / 255)),
    (2, (77 / 255, 175 / 255, 74 / 255)),
    (3, (152 / 255, 78 / 255, 163 / 255)),
    (4, (255 / 255, 127 / 255, 0 / 255)),
    (5, (255 / 255, 255 / 255, 51 / 255)),
    (6, (166 / 255, 86 / 255, 40 / 255)),
    (7, (247 / 255, 129 / 255, 191 / 255)),
    (8, (153 / 255, 153 / 255, 153 / 255)),
)

set2 = Colormap(
    (0, (102 / 255, 194 / 255, 165 / 255)),
    (1, (252 / 255, 141 / 255, 98 / 255)),
    (2, (141 / 255, 160 / 255, 203 / 255)),
    (3, (231 / 255, 138 / 255, 195 / 255)),
    (4, (166 / 255, 216 / 255, 84 / 255)),
    (5, (255 / 255, 217 / 255, 47 / 255)),
    (6, (229 / 255, 196 / 255, 148 / 255)),
    (7, (179 / 255, 179 / 255, 179 / 255)),
)

set3 = Colormap(
    (0, (141 / 255, 211 / 255, 199 / 255)),
    (1, (255 / 255, 255 / 255, 179 / 255)),
    (2, (190 / 255, 186 / 255, 218 / 255)),
    (3, (251 / 255, 128 / 255, 114 / 255)),
    (4, (128 / 255, 177 / 255, 211 / 255)),
    (5, (253 / 255, 180 / 255, 98 / 255)),
    (6, (179 / 255, 222 / 255, 105 / 255)),
    (7, (252 / 255, 205 / 255, 229 / 255)),
    (8, (217 / 255, 217 / 255, 217 / 255)),
    (9, (188 / 255, 128 / 255, 189 / 255)),
    (10, (204 / 255, 235 / 255, 197 / 255)),
    (11, (255 / 255, 237 / 255, 111 / 255)),
)

qualitative_colormaps = {
    "accent": accent,
    "dark2": dark2,
    "paired": paired,
    "pastel1": pastel1,
    "pastel2": pastel2,
    "set1": set1,
    "set2": set2,
    "set3": set3,
}


def colorbar(height, length, colormap, category=False):
    """Return the channels of a colorbar."""
    cbar = np.tile(np.arange(length) * 1.0 / (length - 1), (height, 1))
    cmin = colormap.values.min()
    cmax = colormap.values.max()
    crange = (cmax - cmin)
    if category:
        # add an extra buffer around colormap limits to show full category
        cbar = cbar * (crange + 1) + (cmin - 0.5)
        cbar = np.round(cbar)
    else:
        cbar = (cbar * crange) + colormap.values.min()

    return colormap.colorize(cbar)


def palettebar(height, length, colormap):
    """Return the channels of a palettebar."""
    cbar = np.tile(np.arange(length) * 1.0 / (length - 1), (height, 1))
    cbar = (cbar * (colormap.values.max() + 1 - colormap.values.min())
            + colormap.values.min())

    return colormap.palettize(cbar)
